"""Tools that require the internet, and get asteroids"""
import os
import pickle
import warnings
from dataclasses import dataclass
from glob import glob
from typing import Optional

import astropy.units as u
import fitsio
import lightkurve as lk
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tess_ephem as te
from astropy.stats import sigma_clip
from astropy.time import Time
from astropy.utils.exceptions import AstropyWarning
from matplotlib import patches
from scipy.interpolate import interp1d
from scipy.ndimage import label
from tqdm import tqdm

from . import PACKAGEDIR
from .scene import StarScene

sector_times = pickle.load(open(f"{PACKAGEDIR}/data/tess_sector_times.pkl", "rb"))


@dataclass
class AsteroidMetaCollection:
    name: str
    sectors: list
    cameras: list
    ccds: list
    aper: Optional = 7

    def __post_init__(self):
        self.asteroids = []
        last_sector = 0
        for sector, camera, ccd in zip(self.sectors, self.cameras, self.ccds):
            if sector not in sector_times.keys():
                continue
            if last_sector != sector:
                t = Time(sector_times[sector] + 2457000, format="jd")
                t += np.median(np.diff(t.value)) / 2
                asteroid = te.ephem(self.name, sector=sector, time=t, verbose=True)
                last_sector = sector
            k = (asteroid.camera == camera) & (asteroid.ccd == ccd)
            am = AsteroidMeta.from_dataframe(
                asteroid[k].join(pd.DataFrame(index=t), how="right"),
                name=self.name,
                aper=self.aper,
                sector=sector,
                camera=camera,
                ccd=ccd,
            )
            if am.mask.any():
                self.asteroids.append(am)
        if len(self.asteroids) == 0:
            raise ValueError(f"No valid asteroid data for sectors {self.sectors}")

    def __getitem__(self, key):
        return self.asteroids[key]

    def __len__(self):
        return len(self.asteroids)

    @staticmethod
    def from_name(name, sector=None, aper=7):
        if sector is not None:
            t = Time(sector_times[sector] + 2457000, format="jd")
            t += np.median(np.diff(t.value)) / 2
            ephem = te.ephem(name, time=t[::12])
        else:
            ephem = te.ephem(name, interpolation_step="6H")
        sectors, cameras, ccds = np.asarray(
            ephem.groupby(["sector", "camera", "ccd"])
            .first()
            .reset_index()[["sector", "camera", "ccd"]]
        ).T
        return AsteroidMetaCollection(name, sectors, cameras, ccds, aper=aper)

    def __repr__(self):
        return f"AsteroidMetaCollection: {self.name} ({len(self)} objects)"


@dataclass
class AsteroidMeta:
    row: list
    col: list
    vmag: list
    time: list
    name: Optional = None
    aper: Optional = 7
    sun_distance: Optional = None
    obs_distance: Optional = None
    phase_angle: Optional = None
    sector: Optional = None
    camera: Optional = None
    ccd: Optional = None

    def __repr__(self):
        repr = f"AsteroidMeta name:{self.name}"
        for key in ["sector", "camera", "ccd"]:
            if getattr(self, key) is not None:
                repr = f"{repr} {key}:{getattr(self, key)}"
        return repr

    @property
    def speed(self):
        if self.mask.any():
            return (
                np.abs(
                    np.gradient(
                        np.asarray(np.hypot(self.row, self.col))[self.mask],
                        self.time[self.mask],
                    )
                )
                * u.pixel
                / u.day
            )
        else:
            return np.atleast_1d(0) * u.pixel / u.day

    @property
    def lag(self):
        return (8 * u.pixel) / self.speed.min()

    @property
    def _fr(self):
        return interp1d(
            self.time[self.mask],
            np.asarray(self.row[self.mask]),
            bounds_error=False,
            fill_value="extrapolate",
        )

    @property
    def _fc(self):
        return interp1d(
            self.time[self.mask],
            np.asarray(self.col[self.mask]),
            bounds_error=False,
            fill_value="extrapolate",
        )

    def __post_init__(self):
        self.mask = (
            (self.row >= self.aper // 2)
            & (self.col >= self.aper // 2)
            & (self.row < (2048 - self.aper // 2))
            & (self.col < (2048 - self.aper // 2))
        )
        if not self.mask.any():
            self.xlag, self.x, self.xlead = (
                (self.row.astype(int), self.col.astype(int)),
                (self.row.astype(int), self.col.astype(int)),
                (self.row.astype(int), self.col.astype(int)),
            )
        else:
            self.xlag, self.x, self.xlead = [
                (
                    self._fr(self.time + l).astype(int),
                    self._fc(self.time + l).astype(int),
                )
                for l in [-self.lag.value, 0, self.lag.value]
            ]
            self.mask &= (
                (self.xlag[0] >= self.aper // 2)
                & (self.xlag[1] >= self.aper // 2)
                & (self.xlag[0] < (2048 - self.aper // 2))
                & (self.xlag[1] < (2048 - self.aper // 2))
            )
            self.mask &= (
                (self.xlead[0] >= self.aper // 2)
                & (self.xlead[1] >= self.aper // 2)
                & (self.xlead[0] < (2048 - self.aper // 2))
                & (self.xlead[1] < (2048 - self.aper // 2))
            )

    @staticmethod
    def from_dataframe(df, name=None, sector=None, camera=None, ccd=None, aper=7):
        return AsteroidMeta(
            row=np.asarray(df.row) - 0.5,
            col=np.asarray(df.column) - 45.5,
            vmag=np.asarray(df.vmag),
            sun_distance=np.asarray(df["sun_distance"]),
            obs_distance=np.asarray(df["obs_distance"]),
            phase_angle=np.asarray(df["phase_angle"]),
            time=np.asarray([i.jd for i in df.index]),
            sector=sector,
            camera=camera,
            ccd=ccd,
            aper=aper,
            name=name,
        )


@dataclass
class AsteroidExtractor:
    """Class to get asteroids out of tess data"""

    dir: str
    sector: int

    def __post_init__(self):
        """Class to get asteroids from TESS data

        Parameters
        ----------
        fnames: list of str
            Filenames of the TESS"""

        if not os.path.isdir(self.dir):
            raise ValueError("No such directory")
        if not os.path.isdir(f"{self.dir}/sector{self.sector:02}"):
            raise ValueError("No such sector")

        self.scenes, self.fnames = {}, {}
        for camera in np.arange(1, 5):
            for ccd in np.arange(1, 5):
                if StarScene.exists(self.sector, camera, ccd):
                    if camera not in self.scenes.keys():
                        self.scenes[camera] = {}
                    self.scenes[camera][ccd] = StarScene.from_disk(
                        self.sector, camera, ccd
                    )
                    if os.path.isdir(
                        f"{self.dir}/sector{self.sector:02}/camera{camera:02}/ccd{ccd:02}"
                    ):
                        if camera not in self.fnames.keys():
                            self.fnames[camera] = {}
                        self.fnames[camera][ccd] = np.sort(
                            glob(
                                f"{self.dir}/sector{self.sector:02}/camera{camera:02}/ccd{ccd:02}/*_ffic.fits"
                            )
                        )
                    else:
                        raise ValueError(
                            f"Please provide fits files for camera {camera} ccd {ccd}"
                        )
        self.time = (
            sector_times[self.sector]
            + np.median(np.diff(sector_times[self.sector])) / 2
        ) + 2457000

    def __repr__(self):
        return f"AsteroidExtractor Sector {self.sector}"

    def get_asteroid_arrays(self, ast, aper=21):
        fnames = self.fnames[ast.camera][ast.ccd]
        ar = np.zeros((len(fnames), aper, aper, 3)) * np.nan
        er = np.zeros((len(fnames), aper, aper, 3)) * np.nan
        for idx, x in enumerate([ast.xlag, ast.x, ast.xlead]):
            r, c = x
            for tdx in range(len(fnames)):
                if (
                    (r[tdx] >= aper // 2)
                    & (c[tdx] >= aper // 2)
                    & (r[tdx] < (2048 + aper // 2 + 1))
                    & (c[tdx] < (2048 + aper // 2 + 1))
                ):
                    ar[tdx, :, :, idx] = fitsio.FITS(fnames[tdx])[1][
                        r[tdx] - aper // 2 : r[tdx] + aper // 2 + 1,
                        44 + c[tdx] - aper // 2 : 44 + c[tdx] + aper // 2 + 1,
                    ]
                    er[tdx, :, :, idx] = fitsio.FITS(fnames[tdx])[2][
                        r[tdx] - aper // 2 : r[tdx] + aper // 2 + 1,
                        44 + c[tdx] - aper // 2 : 44 + c[tdx] + aper // 2 + 1,
                    ]
        return ar, er

    def get_asteroid_lightcurve(self, name, aper=21, plot=False, quality_mask=175):
        collection = AsteroidMetaCollection.from_name(
            name, sector=self.sector, aper=aper
        )
        R, C = np.mgrid[:aper, :aper]
        R -= aper // 2
        C -= aper // 2

        lcs = []
        for ast in collection:
            ast.mask &= (
                self.scenes[ast.camera][ast.ccd].quality.astype(int)
                & (8192 | quality_mask)
            ) == 0

            ar, er = self.get_asteroid_arrays(ast, aper=aper)

            model = np.zeros(ar.shape)
            for idx, x in enumerate([ast.xlag, ast.x, ast.xlead]):
                r, c = x
                row3d = np.zeros((r.shape[0], aper), int)
                row3d = (
                    np.asarray(r).astype(int)
                    + np.arange(-aper // 2 + 1, aper // 2 + 1)[:, None]
                ).T
                column3d = np.zeros((r.shape[0], aper), int)
                column3d = (
                    np.asarray(c).astype(int)
                    + np.arange(-aper // 2 + 1, aper // 2 + 1)[:, None]
                ).T
                model[:, :, :, idx] = self.scenes[ast.camera][ast.ccd].model_moving(
                    row3d, column3d, time_indices=np.where(ast.mask)[0]
                )

            # Get aperture
            aperture_mask = self._get_aperture_mask(
                ar[:, :, :, 1] - model[:, :, :, 1], er[:, :, :, 1]
            )

            bkg = (
                np.nanmedian(
                    ar[:, ~aperture_mask, :] - model[:, ~aperture_mask, :], axis=1
                )
            )[:, None, None, :]
            ar -= bkg
            # Get photometery
            flux = (ar[:, aperture_mask, 1] - model[:, aperture_mask, 1]).sum(axis=1)[
                ast.mask
            ]
            flux_err = (er[:, aperture_mask, 1] ** 2).sum(axis=1)[ast.mask] ** 0.5

            # Interpolate leading/lagging
            l1 = interp1d(
                ast.time[ast.mask] - ast.lag.value,
                (ar[:, aperture_mask, 0] - model[:, aperture_mask, 0]).sum(axis=1)[
                    ast.mask
                ],
                fill_value="extrapolate",
                bounds_error=False,
            )(ast.time[ast.mask])
            l2 = interp1d(
                ast.time[ast.mask] + ast.lag.value,
                (ar[:, aperture_mask, 2] - model[:, aperture_mask, 2]).sum(axis=1)[
                    ast.mask
                ],
                fill_value="extrapolate",
                bounds_error=False,
            )(ast.time[ast.mask])

            # Mask out bad cadences based on leading and lagging
            cadence_mask = np.ones(ast.mask.sum(), bool)
            k = np.isfinite(np.nansum(np.vstack([flux]), axis=0))

            for dat in [l1, l2, np.hypot(l1, l2), l1 / l2]:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=AstropyWarning)
                    m = sigma_clip(dat[k]).mask
                    m[~np.isfinite(dat[k])] = False
                    cadence_mask[k] &= ~m
                    m = sigma_clip(np.gradient(dat[k], ast.time[ast.mask][k])).mask
                    m[~np.isfinite(dat[k])] = False
                    cadence_mask[k] &= ~m

            # These might help users
            xcent = np.average(
                R[aperture_mask][None, :] * np.ones((ar.shape[0], 1)),
                weights=(ar[:, aperture_mask, 1] - model[:, aperture_mask, 1]),
                axis=1,
            )
            ycent = np.average(
                C[aperture_mask][None, :] * np.ones((ar.shape[0], 1)),
                weights=(ar[:, aperture_mask, 1] - model[:, aperture_mask, 1]),
                axis=1,
            )

            flux_corr = 10 ** (
                (ast.vmag[ast.mask] - np.median(ast.vmag[ast.mask])) * -0.4
            )

            # Package
            lc = lk.LightCurve(
                time=self.time[ast.mask][cadence_mask],
                flux=(flux / flux_corr)[cadence_mask],
                flux_err=flux_err[cadence_mask],
                targetid=ast.name,
                meta={
                    "LABEL": ast.name,
                    "MISSION": "TESS",
                    "TELESCOP": "TESS",
                    "SECTOR": ast.sector,
                    "TARGETID": ast.name,
                },
            )

            central_pixel = [aper // 2, aper // 2]
            lc["central_pixel"] = (
                ar[:, central_pixel[0], central_pixel[1], 1]
                - model[:, central_pixel[0], central_pixel[1], 1]
            )[ast.mask][cadence_mask]
            lc["xcentroid"] = xcent[ast.mask][cadence_mask]
            lc["ycentroid"] = ycent[ast.mask][cadence_mask]
            lc["ephem_row"] = ast.row[ast.mask][cadence_mask]
            lc["ephem_col"] = ast.col[ast.mask][cadence_mask]
            lc["vmag"] = ast.vmag[ast.mask][cadence_mask]
            lc["sun_distance"] = ast.sun_distance[ast.mask][cadence_mask]
            lc["obs_distance"] = ast.obs_distance[ast.mask][cadence_mask]
            lc["phase_angle"] = ast.phase_angle[ast.mask][cadence_mask]
            lc["flux_corr"] = flux_corr[cadence_mask]
            lc["camera"] = np.zeros(cadence_mask.sum(), dtype=int) + ast.camera
            lc["ccd"] = np.zeros(cadence_mask.sum(), dtype=int) + ast.ccd
            lc["l1"] = l1[cadence_mask]
            lc["l2"] = l2[cadence_mask]

            lcs.append(self._correct_lc(lc.remove_nans()))
        if len(lcs) == 0:
            return lk.LightCurve(time=[], flux=[])
        lc, thumb = (
            lk.LightCurveCollection(lcs).stitch(lambda x: x),
            (np.nanmedian(ar[:, :, :, 1] - model[:, :, :, 1], axis=0)),
        )
        if plot:
            ax = _plot_asteroid(lc, thumb, aperture_mask)
            return lc, ax
        # return (lc, thumb, aperture_mask, collection, model)
        return lc

    def _correct_lc(self, lc):

        # Remove outliers in the image centroids
        if len(lc.flux) > 10:
            r, c = lc.ephem_row.value % 1, lc.ephem_col.value % 1
            poly = np.vstack(
                [(lc.time.value - np.mean(lc.time.value)) ** idx for idx in range(4)]
            ).T
            A = np.hstack(
                [
                    poly,
                    r[:, None],
                    (r * c)[:, None],
                ]
            )
            xcent_corr = lc.xcentroid.value - A.dot(
                np.linalg.solve(A.T.dot(A), A.T.dot(lc.xcentroid.value))
            )
            A = np.hstack(
                [
                    poly,
                    c[:, None],
                    (r * c)[:, None],
                ]
            )
            ycent_corr = lc.ycentroid.value - A.dot(
                np.linalg.solve(A.T.dot(A), A.T.dot(lc.ycentroid.value))
            )
            cent_corr = np.hypot(
                xcent_corr - xcent_corr.min(), ycent_corr - ycent_corr.min()
            )
            nknots = int((lc.time.value[-1] - lc.time.value[0]) / 0.75)
            if nknots > 3:
                dm = lk.designmatrix.create_spline_matrix(
                    lc.time.value - np.mean(lc.time.value), n_knots=nknots
                )
            else:
                dm = lk.DesignMatrix(poly)
            r = lk.RegressionCorrector(
                lk.LightCurve(time=lc.time.value, flux=cent_corr, targetid="cent")
            )
            _ = r.correct(dm, sigma=3)
            outliers = r.outlier_mask
            outliers |= np.gradient(r.outlier_mask.astype(float)) != 0
            lc = lc[~outliers]
        return lc

    def _get_aperture_mask(self, ar, er, sigma=4):
        e = (np.nansum(er ** 2, axis=0) ** 0.5) / er.shape[0]
        thumb = np.nanmedian(ar, axis=0) / e
        aperture_mask = (
            thumb > np.hstack([thumb[thumb < 0], -thumb[thumb < 0]]).std() * sigma
        )
        if not aperture_mask[ar.shape[1] // 2, ar.shape[2] // 2]:
            raise ValueError("Can not find asteroid")

        # This is taken from lightkurve's create_threshold_mask
        labels = label(aperture_mask)[0]
        # For all pixels above threshold, compute distance to reference pixel:
        label_args = np.argwhere(labels > 0)
        distances = [
            np.hypot(crd[0], crd[1])
            for crd in label_args - np.array([ar.shape[1] // 2, ar.shape[2] // 2])
        ]
        # Which label corresponds to the closest pixel?
        closest_arg = label_args[np.argmin(distances)]
        closest_label = labels[closest_arg[0], closest_arg[1]]
        return labels == closest_label


def get_ffi_times():
    """Gets the times of all the FFI images from MAST by looking at all the file names."""
    from urllib.request import HTTPError

    """Get a dictionary of the times for each TESS FFI and save it as a pkl file"""
    time_dict = {}
    for sector in tqdm(np.arange(1, 300), total=300):
        try:
            df = pd.read_csv(
                f"https://archive.stsci.edu/missions/tess/download_scripts/sector/tesscurl_sector_{sector}_ffir.sh",
                skiprows=1,
                header=None,
            )
        except HTTPError:
            break
        df[["time", "camera", "ccd"]] = np.asarray(
            [
                (int(d[0][20:33]), int(d[0][40]), int(d[0][42]))
                for idx, d in df.iterrows()
            ]
        )
        time_dict[sector] = (
            Time.strptime(
                np.sort(df[(df.camera == 1) & (df.ccd == 1)].time).astype(str),
                "%Y%j%H%M%S",
            ).jd
            + 0.000809
            - 2457000
        )
    pickle.dump(time_dict, open(f"{PACKAGEDIR}/data/tess_sector_times.pkl", "wb"))


def _plot_asteroid(lc, thumbnail, aperture_mask):
    with plt.style.context("seaborn-white"):
        fig = plt.figure(figsize=(15, 3))
        ax = plt.subplot2grid((1, 4), (0, 0), colspan=3, fig=fig)
        lc.remove_outliers(10).errorbar(ax=ax, c="k", ls="-")
        ax.set(xlabel="Time [JD]", ylabel="Flux [counts]")
        ax = plt.subplot2grid((1, 4), (0, 3), fig=fig)
        ax.set(xticks=[], yticks=[], title=lc.label)
        std = np.hstack([thumbnail[thumbnail < 0], -thumbnail[thumbnail < 0]]).std()
        im = ax.imshow(thumbnail / std, vmin=0, vmax=10, cmap="viridis")
        cbar = plt.colorbar(im, ax=ax)
        cbar.set_label("Counts")
        # Overlay the aperture mask if given
        for i in range(thumbnail.shape[0]):
            for j in range(thumbnail.shape[1]):
                if aperture_mask[i, j]:
                    rect = patches.Rectangle(
                        xy=(j - 0.5, i - 0.5),
                        width=1,
                        height=1,
                        color="red",
                        fill=False,
                        hatch="/////",
                        alpha=0.3,
                    )
                    ax.add_patch(rect)
        return ax



"""Class for working with design matrices"""
from abc import ABC, abstractmethod
from copy import deepcopy

from .cupy_numpy_imports import sparse, xp
from .utils import _spline_basis_vector


class design_matrix(ABC):
    """
    Base class to store design matrices. This is an abstract base class.

    `design_matrix` objects have the following properties.

    Attributes
    ----------
    name : str
        Name of the design matrix
    prior_mu : xp.ndarray, int or float
        The prior mean of the design matrix components
    prior_sigma : xp.ndarray, int or float
        The prior standard deviation of the design matrix components
    sigma_f : xp.ndarray
        The weights for each pixel in the design matrix. Default is
        equal weights.
    A : xp.ndarray
        The design matrix array.
    sigma_w_inv: xp.ndarray
            Inverse covariance matrix of the design matrix
    shape : tuple
        Shape of the design matrix
    """

    def __init__(self, name="DM", sigma_f=None, prior_sigma=None, prior_mu=None):
        self.name = name
        self.A = self._build()
        if prior_mu is None:
            self.prior_mu = xp.zeros(self.shape[1])
        else:
            self.prior_mu = prior_mu
        if prior_sigma is None:
            self.prior_sigma = xp.ones(self.shape[1]) * xp.inf
        else:
            self.prior_sigma = prior_sigma
        self._validate()
        self.update_sigma_f(sigma_f)

    def update_sigma_f(self, sigma_f):
        """Update the `sigma_f` values with new ones."""
        if sigma_f is None:
            self.sigma_f = xp.ones(self.shape[0])
        elif sigma_f.ndim == 2:
            self.sigma_f = sigma_f.ravel()
        else:
            self.sigma_f = sigma_f
        if self.A is None:
            return
        if not self.sigma_f.shape[0] == self.shape[0]:
            raise ValueError(f"`sigma_f` must be shape {self.shape[0]}")
        self._build_sigma_w_inv()

    def _validate(self):
        """Check that the priors are the correct shape"""
        if self.A is None:
            return
        if isinstance(self.prior_mu, (int, float)):
            self.prior_mu = xp.ones(self.shape[1]) * self.prior_mu
        else:
            if not self.prior_mu.shape[0] == self.shape[1]:
                raise ValueError(f"`prior_mu` must be shape {self.shape[1]}")
        if isinstance(self.prior_sigma, (int, float)):
            self.prior_sigma = xp.ones(self.shape[1]) * self.prior_sigma
        else:
            if not self.prior_sigma.shape[0] == self.shape[1]:
                raise ValueError(f"`prior_sigma` must be shape {self.shape[1]}")

    def _build_sigma_w_inv(self):
        """Build `sigma_w_inv` and store it so we don't have to remake it."""
        self.AT = self.A.T
        if sparse.issparse(self.A):
            self.sigma_w_inv = self.AT.dot(
                self.A.multiply(sparse.csr_matrix(1 / self.sigma_f[:, None]))
            ) + xp.diag(1 / self.prior_sigma ** 2)
        else:
            self.sigma_w_inv = self.AT.dot(self.A / self.sigma_f[:, None]) + xp.diag(
                1 / self.prior_sigma ** 2
            )

    @abstractmethod
    def _build(self):
        """This method MUST be replaced by subclass"""

        return

    @property
    def shape(self):
        return self.A.shape

    def __repr__(self):
        return f"{self.name} {self.shape}"

    def copy(self):
        """Create a deep copy of `self`"""
        return deepcopy(self)

    def join(self, other):
        """Join two design matrices together, by stacking them horizontally."""
        copy = self.copy()
        if sparse.issparse(copy.A) and sparse.issparse(other.A):
            copy.A = sparse.hstack([copy.A, other.A]).tocsr()
        elif (not sparse.issparse(copy.A)) and (not sparse.issparse(other.A)):
            copy.A = xp.hstack([copy.A, other.A])
        elif sparse.issparse(copy.A) and (not sparse.issparse(other.A)):
            copy.A = sparse.hstack(
                [copy.A, sparse.csr_matrix(deepcopy(other.A))]
            ).tocsr()
        else:
            copy.A = sparse.hstack([sparse.csr_matrix(copy.A), other.A]).tocsr()
        copy.prior_mu = xp.hstack([copy.prior_mu, other.prior_mu])
        copy.prior_sigma = xp.hstack([copy.prior_sigma, other.prior_sigma])
        copy.sigma_f = xp.hypot(copy.sigma_f, other.sigma_f)
        copy._build_sigma_w_inv()
        copy.name = " and ".join([copy.name, other.name])
        return copy

    def __add__(self, other):
        return self.join(other)

    def fit_frame(self, flux):
        """Fit a given frame of data

        Parameters
        ----------
        flux : np.ndarray
            2D array of values, must be shape `(self.shape[1]**0.5, self.shape[1]**0.5)`

        Returns
        -------
        weights : np.ndarray
            1D array of weights, shape `self.shape[0]`, i.e. number of components
        """
        B = (
            self.AT.dot(flux.ravel() / self.sigma_f)
            + self.prior_mu / self.prior_sigma ** 2
        )
        return xp.linalg.solve(self.sigma_w_inv, B)

    def fit_batch(self, flux_cube):
        """Fit a given cube of data

        Parameters
        ----------
        flux : np.ndarray
            3D array of values, must be shape `(N, self.shape[1]**0.5, self.shape[1]**0.5)`

        Returns
        -------
        weights : np.ndarray
            2D array of weights, shape `(N, self.shape[0])`
        """
        B = xp.asarray(
            [
                self.AT.dot(flux.ravel() / self.sigma_f)
                + self.prior_mu / self.prior_sigma ** 2
                for flux in flux_cube
            ]
        ).T
        return xp.linalg.solve(self.sigma_w_inv, B).T

    def dot(self, other):
        """Takes the dot product of self with a vector."""
        return self.A.dot(other)


class TESS_design_matrix(design_matrix):
    """`design_matrix` Class specifically for working with TESS. Note you can not initialize this class, it has no `_build` method.

    This object hard codes certain values, for example that TESS images are 2048x2048 pixels.

    Attributes
    ----------
    ccd : int
        CCD number
    column : None or xp.ndarray
        The column numbers to evaluate the design matrix at. If None, uses all pixels.
    row : None or xp.ndarray
        The column numbers to evaluate the design matrix at. If None, uses all pixels.
    cutout_size : int
            Size of a "cutout" of images to use. Default is 2048. Use a smaller cut out to test functionality
    bore_pixel : tuple
            Location of the pixel closest to the telescope boresight
    name : str
            Name of the design matrix
    prior_mu : xp.ndarray, int or float
            The prior mean of the design matrix components
    prior_sigma : xp.ndarray, int or float
            The prior standard deviation of the design matrix components
    sigma_f : xp.ndarray
            The weights for each pixel in the design matrix. Default is
            equal weights.
    A : xp.ndarray
            The design matrix array.
    sigma_w_inv: xp.ndarray
                    Inverse covariance matrix of the design matrix
    shape : tuple
            Shape of the design matrix
    """

    def __build(self):
        raise ValueError("Can not instantiate a `TESS_design_matrix`")

    def __init__(
        self,
        sigma_f=None,
        prior_sigma=None,
        prior_mu=None,
        ccd=3,
        column=None,
        row=None,
        name="TESS",
        cutout_size=2048,
    ):
        """
        Create a `TESS_design_matrix` object. Note, you can not initialize this object as it has no `_build` method.

        Parameters
        ----------
        sigma_f : xp.ndarray
            The weights for each pixel in the design matrix. Default is
            equal weights.
        prior_sigma : xp.ndarray, int or float
            The prior standard deviation of the design matrix components
        prior_mu : xp.ndarray, int or float
            The prior mean of the design matrix components
        ccd : int
                CCD number
        column : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        row : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        name : str
                Name for design matrix
        cutout_size : int
                Size of a "cutout" of images to use. Default is 2048. Use a smaller cut out to test functionality
        """
        self.cutout_size = cutout_size
        self.ccd = ccd
        if self.ccd in [1, 3]:
            self.bore_pixel = [2048, 2048]
        elif self.ccd in [2, 4]:
            self.bore_pixel = [2048, 0]
        if (column is None) and (row is None):
            row, column = xp.mgrid[: self.cutout_size, : self.cutout_size]
            self.column, self.row = (column - self.bore_pixel[1]) / (2048), (
                row - self.bore_pixel[0]
            ) / (2048)
        elif (column is not None) and (row is not None):
            #             row, column = xp.meshgrid(column, row)

            row, column = xp.meshgrid(row, column)
            row = row.T
            column = column.T
            # Should this be 2048?
            self.column, self.row = (column - self.bore_pixel[1]) / (
                self.cutout_size
            ), (row - self.bore_pixel[0]) / (self.cutout_size)
        else:
            raise ValueError("Specify both column and row")
        super().__init__(
            name=name, sigma_f=sigma_f, prior_sigma=prior_sigma, prior_mu=prior_mu
        )
        self._validate()


class cartesian_design_matrix(TESS_design_matrix):
    """Design matrix for 2D Polynomial in cartesian space"""

    def _build(self):
        A1 = xp.vstack([self.column.ravel() ** idx for idx in range(self.npoly)]).T
        A2 = xp.vstack([self.row.ravel() ** idx for idx in range(self.npoly)]).T
        return xp.hstack(
            [A1 * A2[:, idx][:, None] for idx in xp.arange(0, A2.shape[1])]
        )

    def __init__(
        self,
        sigma_f=None,
        prior_sigma=None,
        prior_mu=None,
        ccd=3,
        npoly=5,
        column=None,
        row=None,
        cutout_size=2048,
    ):
        """
        Create a `cartesian_design_matrix` object.

        Parameters
        ----------
        sigma_f : xp.ndarray
            The weights for each pixel in the design matrix. Default is
            equal weights.
        prior_sigma : xp.ndarray, int or float
            The prior standard deviation of the design matrix components
        prior_mu : xp.ndarray, int or float
            The prior mean of the design matrix components
        ccd : int
                CCD number
        npoly : int
                Order of polymomial to create.
        column : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        row : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        name : str
                Name for design matrix
        cutout_size : int
                Size of a "cutout" of images to use. Default is 2048. Use a smaller cut out to test functionality
        """
        self.npoly = npoly
        super().__init__(
            name="cartesian",
            sigma_f=sigma_f,
            prior_sigma=prior_sigma,
            prior_mu=prior_mu,
            ccd=ccd,
            column=column,
            row=row,
            cutout_size=cutout_size,
        )


class radial_design_matrix(TESS_design_matrix):
    """Design matrix for creating a polynomial in radial dimension from boresight pixel"""

    def _build(self):
        self.rad = xp.hypot(self.column, self.row).ravel() / xp.sqrt(2)
        A = xp.vstack([self.rad.ravel() ** idx for idx in range(self.npoly)]).T
        return A

    def __init__(
        self,
        sigma_f=None,
        prior_sigma=None,
        prior_mu=None,
        ccd=3,
        npoly=10,
        column=None,
        row=None,
        cutout_size=2048,
    ):
        """
        Create a `radial_design_matrix` object.

        Parameters
        ----------
        sigma_f : xp.ndarray
            The weights for each pixel in the design matrix. Default is
            equal weights.
        prior_sigma : xp.ndarray, int or float
            The prior standard deviation of the design matrix components
        prior_mu : xp.ndarray, int or float
            The prior mean of the design matrix components
        ccd : int
                CCD number
        npoly : int
                Order of polymomial to create.
        column : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        row : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        name : str
                Name for design matrix
        cutout_size : int
                Size of a "cutout" of images to use. Default is 2048. Use a smaller cut out to test functionality
        """
        self.npoly = npoly
        super().__init__(
            name="radial",
            sigma_f=sigma_f,
            prior_sigma=prior_sigma,
            prior_mu=prior_mu,
            ccd=ccd,
            column=column,
            row=row,
            cutout_size=cutout_size,
        )


class radial_spline_design_matrix(TESS_design_matrix):
    """Design matrix for creating a polynomial in radial dimension from boresight pixel"""

    def _build(self):
        self.rad = (xp.hypot(self.column, self.row).ravel() / xp.sqrt(2)).ravel()
        knots = (
            xp.linspace(0, 1, self.nknots) ** 0.25 + 1e-10
        )  # This stops numerical instabilities where x==knot value
        knots_wbounds = xp.append(xp.append([0] * (2 - 1), knots), [1] * (2 + 2))
        return sparse.vstack(
            [
                sparse.csr_matrix(_spline_basis_vector(self.rad, 2, i, knots_wbounds))
                for i in xp.arange(-1, len(knots_wbounds) - 2 - 3)
            ]
        ).T

    def __init__(
        self,
        sigma_f=None,
        prior_sigma=None,
        prior_mu=None,
        ccd=3,
        nknots=60,
        column=None,
        row=None,
        cutout_size=2048,
    ):
        """
        Create a `radial_design_matrix` object.

        Parameters
        ----------
        sigma_f : xp.ndarray
            The weights for each pixel in the design matrix. Default is
            equal weights.
        prior_sigma : xp.ndarray, int or float
            The prior standard deviation of the design matrix components
        prior_mu : xp.ndarray, int or float
            The prior mean of the design matrix components
        ccd : int
                CCD number
        nknots : int
                Number of knots
        column : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        row : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        name : str
                Name for design matrix
        cutout_size : int
                Size of a "cutout" of images to use. Default is 2048. Use a smaller cut out to test functionality
        """
        self.nknots = nknots
        super().__init__(
            name="radial",
            sigma_f=sigma_f,
            prior_sigma=prior_sigma,
            prior_mu=prior_mu,
            ccd=ccd,
            column=column,
            row=row,
            cutout_size=cutout_size,
        )


class strap_design_matrix(TESS_design_matrix):
    """Design matrix for creating column-wise offsets for TESS straps"""

    def _build(self):
        d = sparse.csr_matrix(xp.diag(xp.ones(self.column.shape[1])))
        # return sparse.hstack([d] * self.column.shape[0]).T.tocsr()
        A = sparse.hstack(
            [
                sparse.hstack([d] * self.row[:, 0] ** idx).T.tocsr()
                for idx in range(self.npoly)
            ]
        )
        return A

    def __init__(
        self,
        sigma_f=None,
        prior_sigma=None,
        prior_mu=None,
        ccd=3,
        column=None,
        row=None,
        cutout_size=2048,
        npoly=2,
    ):
        """
        Create a `strap_design_matrix` object.

        Parameters
        ----------
        sigma_f : xp.ndarray
            The weights for each pixel in the design matrix. Default is
            equal weights.
        prior_sigma : xp.ndarray, int or float
            The prior standard deviation of the design matrix components
        prior_mu : xp.ndarray, int or float
            The prior mean of the design matrix components
        ccd : int
            CCD number
        column : None or xp.ndarray
            The column numbers to evaluate the design matrix at. If None, uses all pixels.
        row : None or xp.ndarray
            The column numbers to evaluate the design matrix at. If None, uses all pixels.
        name : str
            Name for design matrix
        cutout_size : int
            Size of a "cutout" of images to use. Default is 2048. Use a smaller cut out to test functionality
        npoly : int
            Polynomial order for the strap model
        """
        self.npoly = npoly
        super().__init__(
            name="strap",
            sigma_f=sigma_f,
            prior_sigma=prior_sigma,
            prior_mu=prior_mu,
            ccd=ccd,
            column=column,
            row=row,
            cutout_size=cutout_size,
        )


class spline_design_matrix(TESS_design_matrix):
    """Design matrix for creating a 2D cartesian spline."""

    def _build(self):
        """Builds a 2048**2 x N matrix"""
        x = (self.column[0] * self.cutout_size + self.bore_pixel[1]) / self.cutout_size
        knots = (
            xp.linspace(0, 1, self.nknots) + 1e-10
        )  # This stops numerical instabilities where x==knot value
        knots_wbounds = xp.append(
            xp.append([0] * (self.degree - 1), knots), [1] * (self.degree + 2)
        )

        # 2D sparse matrix, for 2048 pixels
        As = sparse.vstack(
            [
                sparse.csr_matrix(
                    _spline_basis_vector(x, self.degree, i, knots_wbounds)
                )
                for i in xp.arange(-1, len(knots_wbounds) - self.degree - 3)
            ]
        ).T

        # 2D sparse matrix, for 2048 pixels x 2048 columns
        A1 = sparse.vstack([As] * self.column.shape[0]).tocsr()

        # 2D sparse matrix, for 2048 pixels x 2048 rows
        if self.column.shape == (self.cutout_size, self.cutout_size):
            A2 = sparse.vstack(
                [A1[idx :: self.cutout_size] for idx in range(self.cutout_size)]
            ).tocsr()
        else:
            x = (
                self.row[:, 0] * self.cutout_size + self.bore_pixel[0]
            ) / self.cutout_size
            As = sparse.vstack(
                [
                    sparse.csr_matrix(
                        _spline_basis_vector(x, self.degree, i, knots_wbounds)
                    )
                    for i in xp.arange(-1, len(knots_wbounds) - self.degree - 3)
                ]
            ).T
            A2 = sparse.vstack([As] * self.column.shape[1]).tocsr()
            A2 = sparse.vstack(
                [A2[idx :: self.column.shape[0]] for idx in range(self.column.shape[0])]
            ).tocsr()
        return sparse.hstack(
            [A1.multiply(A2[:, idx]) for idx in range(A2.shape[1])]
        ).tocsr()

    def __init__(
        self,
        sigma_f=None,
        prior_sigma=None,
        prior_mu=None,
        ccd=3,
        nknots=40,
        degree=2,
        column=None,
        row=None,
        cutout_size=2048,
    ):
        """
        Create a `spline_design_matrix` object.

        Parameters
        ----------
        sigma_f : xp.ndarray
            The weights for each pixel in the design matrix. Default is
            equal weights.
        prior_sigma : xp.ndarray, int or float
            The prior standard deviation of the design matrix components
        prior_mu : xp.ndarray, int or float
            The prior mean of the design matrix components
        ccd : int
                CCD number
        nknots : int
                Number of knots to use
        degree : int
                Degree of basis spline
        column : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        row : None or xp.ndarray
                The column numbers to evaluate the design matrix at. If None, uses all pixels.
        name : str
                Name for design matrix
        cutout_size : int
                Size of a "cutout" of images to use. Default is 2048. Use a smaller cut out to test functionality
        """
        self.degree = degree
        self.nknots = nknots
        super().__init__(
            name="spline",
            sigma_f=sigma_f,
            prior_sigma=prior_sigma,
            prior_mu=prior_mu,
            ccd=ccd,
            column=column,
            row=row,
            cutout_size=cutout_size,
        )



"""Tools to work with TPFs"""
from .cupy_numpy_imports import xp
from .scene import StarScene


def correct_tpf(tpf, jitter=True):
    """Correct a TESS Target Pixel File

    Parameters
    ----------
    tpf: lightkurve.TargetPixelFile
        TESSCut target pixel file.
    """
    ss = StarScene.from_tpf(tpf)
    time_mask = ss.get_tpf_mask(tpf)
    quality_mask = ss.quality_mask(175)
    if jitter:
        model = ss.model(mask_bad_pixels=False)
        model -= ss.background.average_frame
    else:
        model = ss.background.model()
    model -= xp.nanmedian(model, axis=0)
    return (tpf[time_mask] - model)[quality_mask]



"""basic utility functions"""
import warnings

import fitsio
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy.convolution import Gaussian1DKernel, convolve
from astropy.stats import sigma_clip, sigma_clipped_stats
from astropy.time import Time
from astropy.utils.exceptions import AstropyWarning
from fbpca import pca
from matplotlib import animation
from scipy.signal import medfilt
from tqdm import tqdm

from . import PACKAGEDIR
from .cupy_numpy_imports import load_image, xp


def _align_with_tpf(object, tpf):
    """Returns indicies to align a BackDrop object with a tpf

    Parameters
    ----------
    object: scatterbrain.ScatteredLightBackground or scatterbrain.StarScene
        BackDrop object to align
    tpf : lightkurve.TargetPixelFile
        TPF object to align

    Returns
    -------
    indices: xp.ndarray
        Array of indices in the BackDrop that are in the TPF
    tpf_indices: xp.ndarray
        Array of indices in the TPF that are in the BackDrop
    """
    idxs, jdxs = [], []
    for idx, t in enumerate(tpf.time.value):
        k = (object.tstart - t) < 0
        k &= (object.tstop - t) > 0
        if k.sum() == 1:
            idxs.append(idx)
            jdxs.append(np.where(k)[0][0])
    return np.asarray(jdxs), np.asarray(idxs)


def _spline_basis_vector(x, degree, i, knots):
    """Recursive function to create a single spline basis vector for an ixput x,
    for the ith knot.
    See https://en.wikipedia.org/wiki/B-spline for a definition of B-spline
    basis vectors

    NOTE: This is lifted out of the funcs I wrote for lightkurve

    Parameters
    ----------
    x : cp.ndarray
        Ixput x
    degree : int
        Degree of spline to calculate basis for
    i : int
        The index of the knot to calculate the basis for
    knots : cp.ndarray
        Array of all knots
    Returns
    -------
    B : cp.ndarray
        A vector of same length as x containing the spline basis for the ith knot
    """
    if degree == 0:
        B = xp.zeros(len(x))
        B[(x >= knots[i]) & (x <= knots[i + 1])] = 1
    else:
        da = knots[degree + i] - knots[i]
        db = knots[i + degree + 1] - knots[i + 1]
        if (knots[degree + i] - knots[i]) != 0:
            alpha1 = (x - knots[i]) / da
        else:
            alpha1 = xp.zeros(len(x))
        if (knots[i + degree + 1] - knots[i + 1]) != 0:
            alpha2 = (knots[i + degree + 1] - x) / db
        else:
            alpha2 = xp.zeros(len(x))
        B = (_spline_basis_vector(x, (degree - 1), i, knots)) * (alpha1) + (
            _spline_basis_vector(x, (degree - 1), (i + 1), knots)
        ) * (alpha2)
    return B


def get_star_mask(f):
    """False where stars are. Keep in mind this might be a bad
    set of hard coded parameters for some TESS images!"""
    # This removes pixels where there is a steep flux gradient
    star_mask = (xp.hypot(*xp.gradient(f)) < 30) & (f < 9e4)
    # This broadens that mask by one pixel on all sides
    star_mask = (
        ~(xp.asarray(xp.gradient(star_mask.astype(float))) != 0).any(axis=0) & star_mask
    )
    return star_mask


def _find_saturation_column_centers(mask):
    """
    Finds the center point of saturation columns.
    Parameters
    ----------
    mask : xp.ndarray of bools
        Mask where True indicates a pixel is saturated
    Returns
    -------
    centers : xp.ndarray
        Array of the centers in XY space for all the bleed columns
    """
    centers = []
    radii = []
    idxs = xp.where(mask.any(axis=0))[0]
    for idx in idxs:
        line = mask[:, idx]
        seq = []
        val = line[0]
        jdx = 0
        while jdx <= len(line):
            while line[jdx] == val:
                jdx += 1
                if jdx >= len(line):
                    break
            if jdx >= len(line):
                break
            seq.append(jdx)
            val = line[jdx]
        w = xp.array_split(line, seq)
        v = xp.array_split(xp.arange(len(line)), seq)
        coords = [(idx, v1.mean().astype(int)) for v1, w1 in zip(v, w) if w1.all()]
        rads = [len(v1) / 2 for v1, w1 in zip(v, w) if w1.all()]
        for coord, rad in zip(coords, rads):
            centers.append(coord)
            radii.append(rad)
    centers = xp.asarray(centers)
    radii = xp.asarray(radii)
    return centers, radii


def get_sat_mask(f):
    """False where saturation spikes are. Keep in mind this might be a bad
    set of hard coded parameters for some TESS images!"""
    sat = f > 9e4
    l, r = _find_saturation_column_centers(sat)
    col, row = xp.mgrid[: f.shape[0], : f.shape[1]]
    l, r = l[r > 1], r[r > 1]
    for idx in range(len(r)):
        sat |= (xp.hypot(row - l[idx, 0], col - l[idx, 1]) < (r[idx] * 2)) & (
            xp.abs(col - l[idx, 1]) < xp.ceil(xp.min([r[idx] * 0.5, 7]))
        )
        sat |= xp.hypot(row - l[idx, 0], col - l[idx, 1]) < (r[idx] * 0.75)

    return ~sat


def _package_pca_comps(backdrop, xpca_components=20, split_time_domain=False):
    """Packages the jitter terms into detrending vectors similar to CBVs.
    If `split_time_domain` then splits the jitter into timescales of:
        - t < 0.5 days
        - t > 0.5 days
    Parameters
    ----------
    backdrop: tess_backdrop.FullBackDrop
        Ixput backdrop to package
    xpca_components : int
        Number of pca components to compress into. Default 20, which will result
        in an ntimes x 40 matrix if `split_time_domain`.
    split_time_domain: bool
        If True will split into two different time scales. Otherwise will take a PCA.

    Returns
    -------
    matrix : xp.ndarray
        The packaged jitter matrix will contains the top principle components
        of the jitter matrix.
    """

    for label in ["jitter", "bkg"]:
        comp = getattr(backdrop, label)
        comp = xp.asarray(comp)
        finite = np.isfinite(comp).all(axis=1)
        # If there aren't enough components, just return them.
        if comp.shape[0] < 40:
            setattr(backdrop, label + "_pack", comp)
            continue
        if finite.sum() < 50:
            setattr(backdrop, label + "_pack", comp)
            continue

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=AstropyWarning)
            if label == "jitter":
                sigma = 2
            else:
                sigma = 5
            perc = np.diff(np.nanpercentile(comp, [90, 10], axis=0), axis=0)[0]
            m = ~sigma_clip(perc, sigma=sigma).mask
            comp = comp[:, m]

        # We split at data downlinks where there is a gap of at least 0.2 days
        breaks = xp.where(xp.diff(backdrop.tstart[finite]) > 0.2)[0] + 1
        breaks = xp.hstack([0, breaks, len(backdrop.tstart[finite])])

        comp_short = comp[finite].copy()

        if split_time_domain:
            nb = int(0.5 / xp.median(xp.diff(backdrop.tstart)))
            nb = [nb if (nb % 2) == 1 else nb + 1][0]

            def smooth(x):
                return xp.asarray([medfilt(x[:, tdx], nb) for tdx in range(x.shape[1])])

            comp_medium = xp.hstack(
                [smooth(comp[finite][x1:x2]) for x1, x2 in zip(breaks[:-1], breaks[1:])]
            ).T

            U1, s, V = pca(
                comp_short - comp_medium, xpca_components, n_iter=10, raw=True
            )
            U2, s, V = pca(comp_medium, xpca_components, n_iter=10, raw=True)

            X = xp.hstack(
                [
                    U1,
                    U2,
                ]
            )
            X = xp.hstack(
                [X[:, idx::xpca_components] for idx in range(xpca_components)]
            )
            Xall = np.zeros((backdrop.tstart.shape[0], X.shape[1]))
            Xall[finite] = X
        else:
            X, s, V = pca(comp_short, xpca_components, n_iter=10, raw=True)
            Xall = np.zeros((backdrop.tstart.shape[0], X.shape[1]))
            Xall[finite] = X
        setattr(backdrop, label + "_pack", Xall)
    return


def movie(data, out="out.mp4", scale="linear", title="", **kwargs):
    fig, ax = plt.subplots(1, 1, figsize=(4.5, 4.5))
    ax.set_facecolor("#ecf0f1")
    im = ax.imshow(data[0], origin="lower", **kwargs)
    xlims, ylims = ax.get_xlim(), ax.get_ylim()

    ax.set(xlim=xlims, ylim=ylims)
    ax.set_xticks([])
    ax.set_yticks([])

    def animate(i):
        im.set_array(data[i])
        return im

    anim = animation.FuncAnimation(fig, animate, frames=len(data), interval=30)
    anim.save(out, dpi=150)


def test_strip(fname, value=False):
    """Test whether any of the CCD strips are saturated"""
    f = np.median(
        np.abs(fitsio.FITS(fname)[1][:10, 44 : 2048 + 44].mean(axis=0)).reshape(
            (4, 512)
        ),
        axis=1,
    )
    if value:
        return f
    return f > 10000


def identify_bad_frames(fnames):
    test = np.asarray([test_strip(fname, value=True) for fname in fnames])
    s = sigma_clipped_stats(test.std(axis=1))
    return convolve(((test.std(axis=1) - s[1]) / s[2]) > 8, Gaussian1DKernel(2)) != 0


def minmax(x, shape=2048):
    return np.min(
        [np.max([x, np.zeros_like(x)], axis=0), np.zeros_like(x) + shape - 1], axis=0
    ).astype(int)


def _validate_inputs(fnames, sector, camera, ccd):
    if not isinstance(fnames, (list, xp.ndarray)):
        raise ValueError("Pass an array of file names")
    if not isinstance(fnames[0], (str)):
        raise ValueError("Pass an array of strings")
    if sector is None:
        try:
            sector = int(fnames[0].split("-s")[1].split("-")[0])
        except ValueError:
            raise ValueError("Can not parse file name for sector number")
    if camera is None:
        try:
            camera = fitsio.read_header(fnames[0], ext=1)["CAMERA"]
        except ValueError:
            raise ValueError("Can not find a camera number")
    if ccd is None:
        try:
            ccd = fitsio.read_header(fnames[0], ext=1)["CCD"]
        except ValueError:
            raise ValueError("Can not find a CCD number")
    return fnames, sector, camera, ccd


def get_asteroid_files(catalog_fname, sectors, magnitude_limit=18):
    """Get files for each sector containing asteroid locations in the image."""
    import os
    import pickle

    import tess_ephem as te

    sector_times = pickle.load(open(f"{PACKAGEDIR}/data/tess_sector_times.pkl", "rb"))

    df_raw = pd.read_csv(catalog_fname, low_memory=False)
    for sector in np.atleast_1d(sectors):
        df = (
            df_raw[
                (df_raw.max_Vmag != 0)
                & (df_raw.sector == sector)
                & (df_raw.max_Vmag <= magnitude_limit)
            ]
            .drop_duplicates("pdes")
            .reset_index(drop=True)
        )
        t = Time(sector_times[sector] + 2457000, format="jd")
        t += np.median(np.diff(t.value)) / 2

        asteroid_df = pd.DataFrame(
            columns=np.hstack(
                [
                    "camera",
                    "ccd",
                    "vmag",
                    [f"{i}r" for i in np.arange(len(t))],
                    [f"{i}c" for i in np.arange(len(t))],
                ]
            ),
            dtype=np.int16,
        )
        names = []
        jdx = 0
        for idx, d in tqdm(df.iterrows(), total=len(df), desc=f"Sector {sector}"):
            ast = te.ephem(
                d.pdes, interpolation_step="6H", time=t, sector=sector, verbose=True
            )[["sector", "camera", "ccd", "column", "row", "vmag"]]

            ast.replace(np.nan, -1, inplace=True)
            for camera in ast.camera.unique():
                for ccd in ast[ast.camera == camera].ccd.unique():
                    j = np.asarray((ast.camera == camera) & (ast.ccd == ccd))
                    k = np.in1d(t.jd, [i.jd for i in ast[j].index])
                    row, col = np.zeros((2, k.shape[0])) - 1
                    row[k] = ast[j].row
                    col[k] = ast[j].column
                    names.append(d.pdes)
                    if (ast["vmag"] > 0).sum() == 0:
                        vmagmean = -99
                    else:
                        vmagmean = np.round(ast["vmag"][ast["vmag"] > 0].mean())
                    asteroid_df.loc[jdx] = np.hstack(
                        [camera, ccd, vmagmean, row, col]
                    ).astype(np.int16)
                    jdx += 1
        path = f"{PACKAGEDIR}/data/sector{sector:03}/"
        if not os.path.isdir(path):
            os.mkdir(path)
        if os.path.isfile(f"{path}bright_asteroids.hdf"):
            os.remove(f"{path}bright_asteroids.hdf")
        asteroid_df.to_hdf(
            f"{path}bright_asteroids.hdf",
            **{"key": f"asteroid_sector{sector}", "format": "fixed", "complevel": 9},
        )


def get_asteroid_locations(sector=1, camera=1, ccd=1, times=None):
    """Get the row and column positions of asteroids in sector, camera, ccd

    Returns
    -------

    row:

    col:
    """
    import pickle

    sector_times = pickle.load(open(f"{PACKAGEDIR}/data/tess_sector_times.pkl", "rb"))

    df = pd.read_hdf(f"{PACKAGEDIR}/data/sector{sector:03}/bright_asteroids.hdf")
    df = df[(df.camera == camera) & (df.ccd == ccd)].reset_index(drop=True)

    vmag = np.asarray(df["vmag"])
    row = (
        np.asarray(df)[:, 2:][:, np.asarray([d.endswith("r") for d in df.columns[2:]])]
        - 0.5
    )
    col = (
        np.asarray(df)[:, 2:][:, np.asarray([d.endswith("c") for d in df.columns[2:]])]
        - 45.5
    )
    if times is None:
        time_mask = np.ones(row.shape[1], bool)
    else:
        time_mask = np.any(
            [np.isclose(sector_times[sector], t, atol=1e-6) for t in times], axis=0
        )

    return vmag, row[:, time_mask], col[:, time_mask]


def get_asteroid_mask(sector=1, camera=1, ccd=1, cutout_size=2048, times=None):
    """Load a saved bright asteroid file as a 2048x2048 pixel mask.

    Use time mask to specify which times to use.
    """

    mask = np.zeros((cutout_size, cutout_size), bool)

    vmag, row, col = get_asteroid_locations(
        sector=sector, camera=camera, ccd=ccd, times=times
    )

    def func(row, col, ap=3):
        X, Y = np.mgrid[-ap : ap + 1, -ap : ap + 1]
        aper = np.hypot(X, Y) <= ap
        aper_locs = np.asarray(np.where(aper)).T - ap
        for idx in range(row.shape[0]):
            k = (
                (row[idx] >= 0)
                & (col[idx] > 0)
                & (row[idx] < cutout_size)
                & (col[idx] < cutout_size)
            )
            for loc in aper_locs:
                l1 = minmax(row[idx, k] + loc[0], shape=cutout_size)
                l2 = minmax(col[idx, k] + loc[1], shape=cutout_size)
                mask[l1, l2] = True

    func(row[vmag >= 14], col[vmag >= 14], ap=5)
    func(row[(vmag < 14) & (vmag >= 11)], col[(vmag < 14) & (vmag >= 11)], ap=7)
    func(row[(vmag < 11)], col[(vmag < 11)], ap=9)
    return mask


def get_locs(im_size=2048, batch_size=512):
    """Get an array of the corners of an image if split into batches

    Parameters
    ----------
    im_size : int
        Size of the image to split
    batch_size: int
        Size of the desired batches

    Returns
    -------
    locs: list of list of tuples
        List of batches. Length number of batches. Each element is a list, of tuples
        which describes the corner of the batch.
    """
    locs = []
    nbatches = im_size // batch_size
    if im_size % batch_size != 0:
        nbatches += 1
    for bdx1 in range(nbatches):
        for bdx2 in range(nbatches):
            locs.append(
                [
                    (batch_size * bdx1, np.min([im_size, batch_size * (bdx1 + 1)])),
                    (batch_size * bdx2, np.min([im_size, batch_size * (bdx2 + 1)])),
                ]
            )
    return locs


def get_min_image_from_filenames(fnames, cutout_size=2048):
    """Find the minimum image in a list of file names.

    Breaks image into batches and loads segments of the image so as not to
    fill up memory.

    Parameters
    ----------
    fnames : list of str
        List of filenames
    cutout_size: int, Optional
        Optional size of the image to cut out.
    Returns
    -------
    ar : np.ndarray
        2D image that is the minimum image in the stack.
    """
    batch_size = np.max([512, int(2 ** (np.log2(cutout_size) - 2))])
    f = np.zeros((cutout_size, cutout_size))
    for loc in get_locs(cutout_size, batch_size):
        f[loc[0][0] : loc[0][1], loc[1][0] : loc[1][1]] = np.min(
            [load_image(fname, loc=loc) for fname in fnames],
            axis=0,
        )
    return f



import os
from copy import deepcopy
from dataclasses import dataclass
from typing import Optional

import fitsio
import numpy as np
from astropy.io import fits
from astropy.stats import sigma_clip
from tqdm import tqdm

from . import PACKAGEDIR
from .background import ScatteredLightBackground
from .cupy_numpy_imports import load_image, xp
from .utils import (
    _align_with_tpf,
    _spline_basis_vector,
    _validate_inputs,
    get_asteroid_locations,
    get_locs,
    minmax,
)
from .version import __version__


@dataclass
class StarScene:
    """Class to remove stars and jitter noise from TESS images.

    This class will take a stack of TESS FFIs, and will use the ScatteredLightBackground
    class to remove scattered light from the images. The ScatteredLightBackground also
    provides an estimate of the "jitter" in pixels as the top principal components of
    bright pixels in the dataset, and the average image. StarScene takes these components
    and fits each individual pixel time series



    Parameters
    ----------
    sector : int
        TESS sector number
    camera: int
        TESS camera number
    ccd : int
        TESS CCD number
    batch_size: int
        This will process the fits images in "batches". The larger the batch_size,
        the more memory fitting will take. 512 is a sane default.
    spline: bool
        Whether to correct time with a spline. If False, will use a 2nd order polynomial.
        If True, will use a 40 knot spline for every pixel.
    ncomps: int
        Number of components from the background model to use. Default is 15. Maximum is 20.
    """

    sector: int
    camera: int
    ccd: int
    row: Optional = None
    column: Optional = None
    batch_size: int = 512
    spline: bool = False
    ncomps: int = 10
    dir: Optional = None

    def __post_init__(self):
        if (self.row is None) | (self.column is None):
            self.row = np.arange(2048)
            self.column = np.arange(2048)
        else:
            self.row = np.atleast_1d(self.row)
            self.column = np.atleast_1d(self.column)

        self.ncomps = np.min([self.ncomps, 20])
        self.background = ScatteredLightBackground.from_disk(
            sector=self.sector,
            camera=self.camera,
            ccd=self.ccd,
            row=np.asarray([1]),
            column=np.asarray([1]),
            dir=self.dir,
        )
        self.break_point = np.where(
            (
                np.diff(self.background.tstart)
                / np.median(np.diff(self.background.tstart))
            )
            > 10
        )[0][0]
        self.orbit_masks = [
            np.arange(len(self.background.tstart)) <= self.break_point,
            np.arange(len(self.background.tstart)) > self.break_point,
        ]
        if self.spline:
            self.Xs = [
                self._get_spline_design_matrix(
                    self.orbit_masks[0], bounds=(-0.5, 0), ncomps=self.ncomps
                ),
                self._get_spline_design_matrix(
                    self.orbit_masks[1], bounds=(0, 0.5), ncomps=self.ncomps
                ),
            ]
        else:
            self.Xs = [
                self._get_design_matrix(self.orbit_masks[0], ncomps=self.ncomps),
                self._get_design_matrix(self.orbit_masks[1], ncomps=self.ncomps),
            ]

        self.weights = [
            np.zeros(
                (self.Xs[0].shape[1], self.row.shape[0], self.column.shape[0]),
                np.float32,
            ),
            np.zeros(
                (self.Xs[0].shape[1], self.row.shape[0], self.column.shape[0]),
                np.float32,
            ),
        ]

        self.bad_pixels = [
            np.zeros(
                (self.row.shape[0], self.column.shape[0]),
                bool,
            ),
            np.zeros(
                (self.row.shape[0], self.column.shape[0]),
                bool,
            ),
        ]
        self.tstart = self.background.tstart
        self.tstop = self.background.tstop
        self.quality = self.background.quality
        return

    def quality_mask(self, quality_bitmask=175):
        return (self.quality.astype(int) & (8192 | quality_bitmask)) == 0

    @property
    def empty(self):
        return (self.weights[0] == 0).all() & (self.weights[1] == 0).all()

    @property
    def shape(self):
        return (len(self.tstart), self.row.shape[0], self.column.shape[0])

    def _load_background(self, row, column):
        try:
            self.background = ScatteredLightBackground.from_disk(
                sector=self.sector,
                camera=self.camera,
                ccd=self.ccd,
                row=row,
                column=column,
            )
        except ValueError:
            raise ValueError("No background solutions exist on disk.")

    def __repr__(self):
        return f"StarScene {self.shape}, Sector {self.sector}, Camera {self.camera}, CCD {self.ccd}"

    @property
    def path(self):
        return (
            f"{PACKAGEDIR}/data/sector{self.sector:03}/camera{self.camera:02}/ccd{self.ccd:02}/"
            f"tessstarscene_sector{self.sector}_camera{self.camera}_ccd{self.ccd}.fits"
        )

    def __getitem__(self, key):
        """Set indexing"""
        if self.tstart is None:
            raise ValueError("Can not slice an object with no `tstart` array.")
        attrs = [
            item
            for item in self.__dir__()
            if not item.startswith("_")
            and isinstance(getattr(self, item), (xp.ndarray, list))
            and (xp.asarray(getattr(self, item)).shape[0] == len(self.tstart))
        ]
        copy = deepcopy(self)
        for attr in attrs:
            setattr(copy, attr, np.asarray(getattr(self, attr))[key])
            if isinstance(getattr(self, attr), list):
                setattr(copy, attr, list(getattr(copy, attr)))
        return copy

    def _get_spline_design_matrix(
        self, mask=None, ncomps=20, degree=2, nknots=20, bounds=(-0.5, 0.5)
    ):
        """Construct a design matrix from the ScatteredLightBackground object"""
        if mask is None:
            mask = np.ones(self.background.tstart.shape[0], bool)
        t = (self.background.tstart[mask] - self.background.tstart.mean()) / (
            self.background.tstart[-1] - self.background.tstart[0]
        )
        knots = (
            xp.linspace(*bounds, nknots + 2)[1:-1] + 1e-10
        )  # This stops numerical instabilities where x==knot value
        knots_wbounds = xp.append(
            xp.append([knots[0]] * (degree - 1), knots), [knots[-1]] * (degree + 2)
        )

        # 2D sparse matrix, for 2048 pixels
        As = xp.vstack(
            [
                _spline_basis_vector(t, degree, i, knots_wbounds)
                for i in xp.arange(-1, len(knots_wbounds) - degree - 3)
            ]
        ).T

        bkg = self.background.jitter_pack[mask, :ncomps]
        jitter = self.background.jitter_pack[mask, :ncomps]

        X = np.hstack(
            [
                bkg,
                jitter,
                # np.vstack([t ** 0, t, t ** 2]).T,
                As,
            ]
        )
        return X

    def _get_design_matrix(self, mask=None, ncomps=10):
        """Construct a design matrix from the ScatteredLightBackground object"""
        if mask is None:
            mask = np.ones(self.background.tstart.shape[0], bool)
        t = (self.background.tstart[mask] - self.background.tstart.mean()) / (
            self.background.tstart[-1] - self.background.tstart[0]
        )
        bkg = self.background.jitter_pack[mask, :ncomps]
        jitter = self.background.jitter_pack[mask, :ncomps]

        X = np.hstack(
            [
                bkg,
                jitter,
                np.vstack([t ** 0, t, t ** 2]).T,
            ]
        )
        return X

    @staticmethod
    def exists(sector=None, camera=None, ccd=None, dir=None):
        for type in ["backdrop", "starscene"]:
            fname = f"tess{type}_sector{sector}_camera{camera}_ccd{ccd}.fits"
            if dir is None:
                dir = f"{PACKAGEDIR}/data/sector{sector:03}/camera{camera:02}/ccd{ccd:02}/"
            if dir != "":
                if not os.path.isdir(dir):
                    return False
            if not os.path.isfile(dir + fname):
                return False
        return True

    @staticmethod
    def from_tess_images(
        fnames, sector=None, camera=None, ccd=None, batch_size=512, plot=False, dir=None
    ):
        """Class to remove stars from TESS images

        Parameters
        ----------
        fnames : list of str
            File names to use to build the object
        sector : int
            TESS sector number
        camera: int
            TESS camera number
        ccd : int
            TESS CCD number
        batch_size: int
            This will process the fits images in "cutouts". The larger the cutout,
            the more memory fitting will take. 512 is a sane default.
        """
        fnames, sector, camera, ccd = _validate_inputs(fnames, sector, camera, ccd)
        self = StarScene(
            sector=sector, camera=camera, ccd=ccd, batch_size=batch_size, dir=dir
        )
        self.fit_model(fnames=fnames, batch_size=batch_size)
        if plot:
            fig = self.diagnose(fnames)
            return self, fig
        return self

    def diagnose(self, fnames, npix=512, ntimes=6):
        import matplotlib.pyplot as plt

        row = np.arange(1024 - npix // 2, 1024 + npix // 2)
        column = np.arange(1024 - npix // 2, 1024 + npix // 2)
        tdxs = np.linspace(0, self.shape[0], ntimes + 2, dtype=int)[1:-1]
        fig, ax = plt.subplots(ntimes, 3, figsize=(4 * 3, 4 * ntimes))
        for jdx, tdx in enumerate(tdxs):
            while self.quality[tdx] != 0:
                tdx += 1
            dat = load_image(fnames[tdx])
            v = np.nanpercentile(dat, (5, 95))
            ax[jdx, 0].imshow(dat[row][:, column], vmin=v[0], vmax=v[1])
            ax[jdx, 0].set_ylabel(f"{self.tstart[tdx]} BTJD")
            ax[jdx, 1].imshow(
                self.model(row, column, time_indices=tdx)[0], vmin=v[0], vmax=v[1]
            )
            ax[jdx, 2].imshow(
                dat[row][:, column] - self.model(row, column, time_indices=tdx)[0],
                vmin=-5,
                vmax=5,
            )
            for idx, title in enumerate(["Data", "Model", "Residuals"]):
                if jdx == 0:
                    ax[jdx, idx].set(title=title)
                ax[jdx, idx].set(xticks=[], yticks=[])
        plt.tight_layout()
        return fig

    @property
    def locs(self):
        """The locations of each of the cutouts. List of lists. Format is

        [[mininmum row, maxmimum row], [minimum column, maximum column]]
        """
        return get_locs(2048, 512)

    def load(self, input, dir=None):
        """
        Load a model fit from the data directory.

        Parameters
        ----------
        input: tuple or string
            Either pass a tuple with `(sector, camera, ccd)` or pass
            a file name in `dir` to load
        dir : str
            Optional tring with the directory name

        Returns
        -------
        self: `scatterbrain.StarScene` object
        """
        if isinstance(input, tuple):
            if len(input) == 3:
                sector, camera, ccd = input
                fname = f"tessstarscene_sector{sector}_camera{camera}_ccd{ccd}.fits"
            else:
                raise ValueError("Please pass tuple as `(sector, camera, ccd)`")
        elif isinstance(input, str):
            fname = input
        else:
            raise ValueError("Can not parse input")
        if dir is None:
            dir = f"{PACKAGEDIR}/data/sector{sector:03}/camera{camera:02}/ccd{ccd:02}/"
        if dir != "":
            if not os.path.isdir(dir):
                raise ValueError("No solutions exist")

        with fits.open(dir + fname, lazy_load_hdus=True) as hdu:
            for key in [
                "sector",
                "camera",
                "ccd",
            ]:
                setattr(self, key, hdu[0].header[key])
            setattr(self, "batch_size", hdu[0].header["BATSIZE"])
            if "NCOMPS" in hdu[0].header:
                setattr(self, "ncomps", hdu[0].header["NCOMPS"])
            else:
                setattr(self, "ncomps", 15)

        self.weight_row = np.arange(self.row[0], self.row[-1] + 1)
        self.weight_column = np.arange(self.column[0], self.column[-1] + 1)
        self.weights = [
            fitsio.FITS(dir + fname)[1][
                :, self.row[0] : self.row[-1] + 1, self.column[0] : self.column[-1] + 1
            ],
            fitsio.FITS(dir + fname)[2][
                :, self.row[0] : self.row[-1] + 1, self.column[0] : self.column[-1] + 1
            ],
        ]
        self.bad_pixels = [
            fitsio.FITS(dir + fname)[3][
                self.row[0] : self.row[-1] + 1, self.column[0] : self.column[-1] + 1
            ].astype(bool),
            fitsio.FITS(dir + fname)[4][
                self.row[0] : self.row[-1] + 1, self.column[0] : self.column[-1] + 1
            ].astype(bool),
        ]
        return self

    @staticmethod
    def from_disk(sector, camera, ccd, dir=None, row=None, column=None):
        return StarScene(
            sector=sector,
            camera=camera,
            ccd=ccd,
            batch_size=512,
            row=row,
            column=column,
        ).load((sector, camera, ccd), dir=dir)

    def _package_weights_hdulist(self):
        hdu0 = self.hdu0
        hdu1 = fits.ImageHDU(np.asarray(self.weights[0]), name="ORBIT1")
        hdu2 = fits.ImageHDU(np.asarray(self.weights[1]), name="ORBIT2")
        hdu3 = fits.ImageHDU(self.bad_pixels[0].astype(int), name="BADPIX1")
        hdu4 = fits.ImageHDU(self.bad_pixels[1].astype(int), name="BADPIX2")
        hdul = fits.HDUList([hdu0, hdu1, hdu2, hdu3, hdu4])
        return hdul

    def save(self, output_dir=None, overwrite=False):
        """Save a StarScene"""
        self.hdu0 = fits.PrimaryHDU()
        self.hdu0.header["ORIGIN"] = "scatterbrain"
        self.hdu0.header["AUTHOR"] = "christina.l.hedges@nasa.gov"
        self.hdu0.header["VERSION"] = __version__
        for key in ["sector", "camera", "ccd"]:
            self.hdu0.header[key] = getattr(self, key)
        self.hdu0.header["BATSIZE"] = getattr(self, "batch_size")
        self.hdu0.header["NCOMPS"] = getattr(self, "ncomps")

        if output_dir is None:
            output_dir = f"{PACKAGEDIR}/data/sector{self.sector:03}/camera{self.camera:02}/ccd{self.ccd:02}/"
            if output_dir != "":
                if not os.path.isdir(output_dir):
                    os.makedirs(output_dir)

        hdul = self._package_weights_hdulist()
        fname = (
            f"tessstarscene_sector{self.sector}_camera{self.camera}_ccd{self.ccd}.fits"
        )
        hdul.writeto(output_dir + fname, overwrite=overwrite)
        return

    def get_images(self, fnames, loc, orbit=1):
        """Load the TESS FFIs, remove the best fitting scattered light model, and
        the best "average" frame, and return as an array.


        Parameters
        ----------
        fnames : list of str
            File names to use to build the object
        loc : list of lists
            Location of the cut out in format [[min row, max row], [min col, max col]]
        orbit: int, [1 or 2]
            Which orbit to get
        """
        if loc != [
            [self.background.row[0], self.background.row[-1] + 1],
            [self.background.column[0], self.background.column[-1] + 1],
        ]:
            self._load_background(row=np.arange(*loc[0]), column=np.arange(*loc[1]))
        y = np.zeros(
            (
                (self.background.quality[self.orbit_masks[orbit - 1]] == 0).sum(),
                *np.diff(loc).ravel(),
            )
        )
        idx = 0
        for tdx in np.where(self.orbit_masks[orbit - 1])[0]:
            if self.background.quality[tdx] != 0:
                continue
            y[idx] = (
                load_image(np.asarray(fnames)[tdx], loc=loc)
                - self.background.model(tdx)
                - self.background.average_frame
            )
            idx += 1
        return y

    def _mask_asteroids(self, flux_cube, loc, orbit):
        """Zero out any asteroids in a flux cube.

        Parameters
        ----------
        flux_cube : np.ndarray
            3D flux object, with shape [time, row, column]
        loc : list of lists
            Location of the cut out in format [[min row, max row], [min col, max col]]
        orbit: int, [1 or 2]
            Which orbit you are processing
        """
        tmask = (self.background.quality == 0) & self.orbit_masks[orbit - 1]
        vmag, row, col = get_asteroid_locations(
            self.sector, self.camera, self.ccd, times=self.tstart[tmask]
        )
        t = np.arange(row.shape[1])

        aps = {"faint": 3, "middle": 3, "bright": 5}
        tests = {
            "faint": lambda x: x > 14,
            "middle": lambda x: (x <= 14) & (x > 11),
            "bright": lambda x: x <= 11,
        }
        for idx in np.arange(row.shape[0]):
            ap = [aps[key] for key, item in tests.items() if item(vmag[idx])][0]
            k = (
                (row[idx] >= loc[0][0])
                & (row[idx] < loc[0][1])
                & (col[idx] >= loc[1][0])
                & (col[idx] < loc[1][1])
            )
            if not k.any():
                continue
            rdx, cdx = row[idx][k] - loc[0][0], col[idx][k] - loc[1][0]
            RDX, CDX = np.array_split(
                np.vstack(
                    [
                        minmax(v + offset, flux_cube.shape[1])
                        for v in [rdx, cdx]
                        for offset in np.arange(-ap, ap + 1)
                    ]
                ),
                2,
            )
            bad_pix = [np.vstack([r, c]) for r in RDX for c in CDX]
            for bdx in range(len(bad_pix)):
                flux_cube[t[k], bad_pix[bdx][0], bad_pix[bdx][1]] = 0

    def _fill_weights_block(self, fnames, loc, iter=False):
        """Process a cutout of the TESS FFIs. Will process the `loc` cutout.

        Parameters
        ----------
        fnames : list of str
            File names to use to build the object
        loc : list of lists
            Location of the cut out in format [[min row, max row], [min col, max col]]
        iter: bool
            Whether to iterate the fit. If True, will repeat the fit, masking
            the largest values in each pixel time series.
        """
        for orbit in [1, 2]:
            y = self.get_images(fnames, loc, orbit=orbit)

            self._mask_asteroids(y, loc=loc, orbit=orbit)
            s = (y.shape[0], np.product(y.shape[1:]))
            X = self.Xs[orbit - 1][
                self.background.quality[self.orbit_masks[orbit - 1]] == 0
            ]
            prior_sigma = np.ones(X.shape[1]) * 1e8
            ws = np.linalg.solve(
                X.T.dot(X) + np.diag(1 / prior_sigma ** 2), X.T.dot(y.reshape(s))
            )
            res = y - X.dot(ws).reshape(y.shape)
            m = sigma_clip(res, sigma_upper=4, sigma_lower=1e10, axis=(1, 2)).mask
            self.bad_pixels[orbit - 1][loc[0][0] : loc[0][1], loc[1][0] : loc[1][1]] = (
                m.sum(axis=0) / y.shape[0]
            ) >= 0.05

            if iter:
                # res[
                #     sigma_clip(
                #         np.abs(res).sum(axis=(1, 2)), sigma_upper=5, sigma_lower=1e10
                #     ).mask
                # ] = 0
                res[
                    self.bad_pixels[orbit - 1][
                        loc[0][0] : loc[0][1], loc[1][0] : loc[1][1]
                    ]
                ] = 0
                for idx in range(5):
                    k = res == np.max(res, axis=0)
                    k = (
                        k
                        | np.vstack([k[1:], k[0][None, :, :]])
                        | np.vstack([k[-1][None, :, :], k[:-1]])
                    )
                    res[k] = 0
                ws2 = np.linalg.solve(X.T.dot(X), X.T.dot((y * (res != 0)).reshape(s)))
                self.weights[orbit - 1][
                    :, loc[0][0] : loc[0][1], loc[1][0] : loc[1][1]
                ] = ws2.reshape((ws.shape[0], *y.shape[1:])).astype(np.float32)
            else:
                self.weights[orbit - 1][
                    :, loc[0][0] : loc[0][1], loc[1][0] : loc[1][1]
                ] = ws.reshape((ws.shape[0], *y.shape[1:])).astype(np.float32)

    def fit_model(self, fnames, batch_size=512):
        """Fit a StarScene model to a set of TESS filenames"""
        for loc in tqdm(self.locs, desc=f"{self.sector}, {self.camera}, {self.ccd} "):
            self._fill_weights_block(fnames=fnames, loc=loc)
        for idx in range(2):
            self.bad_pixels[idx] |= (
                np.asarray(np.gradient(self.bad_pixels[idx].astype(float))) != 0
            ).any(axis=0)

    def model(
        self, row=None, column=None, time_indices=None, mask_bad_pixels=True, trend=True
    ):
        """Returns a model for a given row and column.

        Parameters
        ----------
        row: np.ndarray
            Row to evaluate at
        column: np.ndarray
            Column to evaluate at
        time_indices: None, int, list of int
            Which indices to evaluate the model at. If None will use all.

        Returns
        -------
        model : np.ndarray
            Array of shape (time_indices, row, column) which has the full scene
            model, including scattered light.
        """
        if (row is None) | (column is None):
            row, column = self.row, self.column
        if time_indices is None:
            time_indices = np.arange(self.shape[0])
        single_frame = False
        if isinstance(time_indices, int):
            time_indices = [time_indices]
            single_frame = True
        if (len(row) != len(self.background.row)) | (
            len(column) != len(self.background.column)
        ):
            self._load_background(row=row, column=column)
        else:
            if not (
                (self.background.row == row).all()
                and (self.background.column == column).all()
            ):
                self._load_background(row=row, column=column)
        # r = np.where(np.in1d(self.weight_row, row))[0]
        # c = np.where(np.in1d(self.weight_column, column))[0]
        r, c = row - row[0], column - column[0]
        w1 = self.weights[0][:, r][:, :, c].reshape(
            (self.weights[0].shape[0], row.shape[0] * column.shape[0])
        )
        w2 = self.weights[1][:, r][:, :, c].reshape(
            (self.weights[1].shape[0], row.shape[0] * column.shape[0])
        )
        xmask1 = np.in1d(np.where(self.orbit_masks[0])[0], time_indices)
        xmask2 = np.in1d(np.where(self.orbit_masks[1])[0], time_indices)
        if trend:
            m1, m2 = self.Xs[0][xmask1].dot(w1), self.Xs[1][xmask2].dot(w2)
        else:
            m1, m2 = self.Xs[0][:, : self.ncomps * 2][xmask1].dot(
                w1[: self.ncomps * 2]
            ), self.Xs[1][:, : self.ncomps * 2][xmask2].dot(w2[: self.ncomps * 2])
        if mask_bad_pixels:
            bp1 = self.bad_pixels[0][r][:, c].reshape((row.shape[0] * column.shape[0]))
            bp2 = self.bad_pixels[1][r][:, c].reshape((row.shape[0] * column.shape[0]))
            m1[:, bp1] = np.nan
            m2[:, bp2] = np.nan
        model = np.vstack([m1, m2])
        model = model.reshape((model.shape[0], row.shape[0], column.shape[0]))
        background = self.background.model(time_indices)
        if single_frame:
            return model[0] + self.background.average_frame + background[0]
        return model + self.background.average_frame + background

    def model_moving(self, row3d, column3d, star_model=True, time_indices=None):
        if time_indices is None:
            time_indices = np.arange(self.shape[0])
        if (row3d.shape[0] != self.shape[0]) | (column3d.shape[0] != self.shape[0]):
            raise ValueError("Pass Row and Column positions for all times. ")
        model = np.zeros((row3d.shape[0], row3d.shape[1], column3d.shape[1])) * np.nan
        for orbit in [0, 1]:
            # mask = self.orbit_masks[orbit]
            mask = self.orbit_masks[orbit] & np.in1d(
                np.arange(self.shape[0]), time_indices
            )
            if mask.sum() == 0:
                continue
            s = (self.weights[orbit].shape[0], row3d.shape[1] * column3d.shape[1])
            r1, r2 = row3d[mask].min(), row3d[mask].max() + 1
            c1, c2 = column3d[mask].min(), column3d[mask].max() + 1

            self._load_background(np.arange(r1, r2), np.arange(c1, c2))
            background = self.background.model(np.where(mask)[0])
            for idx, tdx, row, column in zip(
                np.arange(mask.sum()), np.where(mask)[0], row3d[mask], column3d[mask]
            ):

                model[tdx] = (
                    +background[idx][row - r1][:, column - c1]
                    + self.background.average_frame[row - r1][:, column - c1]
                )
                if star_model:
                    w1 = self.weights[orbit][:, row][:, :, column].reshape(s)
                    model[tdx] += (
                        self.Xs[orbit][idx]
                        .dot(w1)
                        .reshape((row.shape[0], column.shape[0]))
                    )
                    model[tdx][self.bad_pixels[orbit][row][:, column]] = np.nan
        return model

    @staticmethod
    def from_tpf(tpf, dir=None):
        scene = StarScene(
            sector=tpf.sector,
            camera=tpf.camera,
            ccd=tpf.ccd,
            row=np.arange(tpf.shape[1]) + tpf.row - 1,
            column=np.arange(tpf.shape[2]) + tpf.column - 44 - 1,
        ).load((tpf.sector, tpf.camera, tpf.ccd), dir=dir)
        idx, jdx = _align_with_tpf(scene, tpf)
        return scene[idx]

    def get_tpf_mask(self, tpf):
        idx, jdx = _align_with_tpf(self, tpf)
        return jdx



import os
from copy import deepcopy

import fitsio
from astropy.io import fits
from tqdm import tqdm

from . import PACKAGEDIR
from .cupy_numpy_imports import load_image, np, xp
from .designmatrix import (
    cartesian_design_matrix,
    radial_spline_design_matrix,
    spline_design_matrix,
    strap_design_matrix,
)
from .utils import (
    _align_with_tpf,
    _package_pca_comps,
    _validate_inputs,
    get_asteroid_mask,
    get_min_image_from_filenames,
    get_sat_mask,
    get_star_mask,
    identify_bad_frames,
)
from .version import __version__


class ScatteredLightBackground(object):
    """Class for working with TESS data to fit and use scattered background models.

    ScatteredLightBackground will automatically set up a reasonable background model for you. If you
    want to tweak the model, check out the `design_matrix` API.
    """

    def __init__(
        self,
        sector,
        camera,
        ccd,
        column=None,
        row=None,
        sigma_f=None,
        nknots=50,
        cutout_size=2048,
        tstart=None,
        tstop=None,
        quality=None,
        verbose=False,
        njitter=10000,
        strap_npoly=2,
    ):
        """Initialize a `ScatteredLightBackground` object either for fitting or loading a model.

        Parameters
        ----------
        sector : int
            Sector number
        camera : int
            Camera number
        ccd : int
            CCD number
        column : None or xp.ndarray
            The column numbers to evaluate the design matrix at. If None, uses all pixels.
        row : None or xp.ndarray
            The column numbers to evaluate the design matrix at. If None, uses all pixels.
        sigma_f : xp.ndarray
            The weights for each pixel in the design matrix. Default is equal weights.
        nknots : int
                Number of knots to for spline matrix
        cutout_size : int
                Size of a "cutout" of images to use. Default is 2048. Use a smaller cut out to test functionality
        """

        self.sector = sector
        self.camera = camera
        self.ccd = ccd
        self.verbose = verbose
        self.njitter = njitter
        self.strap_npoly = strap_npoly
        self.A1 = radial_spline_design_matrix(
            column=column,
            row=row,
            ccd=ccd,
            sigma_f=sigma_f,
            prior_mu=2,
            prior_sigma=100,
            cutout_size=cutout_size,
        ) + cartesian_design_matrix(
            column=column,
            row=row,
            ccd=ccd,
            sigma_f=sigma_f,
            prior_mu=2,
            prior_sigma=100,
            cutout_size=cutout_size,
            npoly=4,
        )

        if column is not None:
            strap_prior_sigma = np.hstack(
                [
                    np.ones(column.shape[0]) * 30 ** (idx + 1)
                    for idx in range(self.strap_npoly)
                ]
            )
        else:
            strap_prior_sigma = np.hstack(
                [
                    np.ones(cutout_size) * 30 ** (idx + 1)
                    for idx in range(self.strap_npoly)
                ]
            )
        self.A2 = (
            spline_design_matrix(
                column=column,
                row=row,
                ccd=ccd,
                sigma_f=sigma_f,
                prior_sigma=200,
                nknots=nknots,
                cutout_size=cutout_size,
            )
            + radial_spline_design_matrix(
                column=column,
                row=row,
                ccd=ccd,
                sigma_f=sigma_f,
                prior_sigma=300,
                cutout_size=cutout_size,
            )
            + strap_design_matrix(
                column=column,
                row=row,
                ccd=ccd,
                sigma_f=sigma_f,
                prior_sigma=strap_prior_sigma,
                cutout_size=cutout_size,
                npoly=self.strap_npoly,
            )
        )
        self.cutout_size = cutout_size
        if row is None:
            self.column = np.arange(self.cutout_size)
        else:
            self.column = column
        if row is None:
            self.row = np.arange(self.cutout_size)
        else:
            self.row = row
        self.weights_basic = []
        self.weights_full = []
        self.jitter = []
        self.bkg = []
        # self._average_frame = xp.zeros(self.shape)
        # self._average_frame_count = 0
        self._average_frame = xp.zeros(self.shape)
        self.tstart = tstart
        self.tstop = tstop
        self.quality = quality

    def update_sigma_f(self, sigma_f):
        self.A1.update_sigma_f(sigma_f)
        self.A2.update_sigma_f(sigma_f)

    def clean(self):
        """Resets the weights and average frame"""
        self.weights_basic = []
        self.weights_full = []
        # self._average_frame = xp.zeros(self.shape)
        # self._average_frame_count = 0
        self._average_frame = xp.zeros(self.shape)
        self.jitter = []
        self.bkg = []

    def __repr__(self):
        return f"ScatteredLightBackground CCD:{self.ccd} ({len(self.weights_basic)} frames)"

    def __getitem__(self, key):
        """Set indexing"""
        if self.tstart is None:
            raise ValueError("Can not slice an object with no `tstart` array.")
        attrs = [
            item
            for item in self.__dir__()
            if not item.startswith("_")
            and isinstance(getattr(self, item), (xp.ndarray, list))
            and (xp.asarray(getattr(self, item)).shape[0] == len(self.tstart))
        ]
        copy = deepcopy(self)
        for attr in attrs:
            setattr(copy, attr, np.asarray(getattr(self, attr))[key])
            if isinstance(getattr(self, attr), list):
                setattr(copy, attr, list(getattr(copy, attr)))
        return copy

    def quality_mask(self, quality_bitmask=175):
        return (self.quality.astype(int) & (8192 | quality_bitmask)) == 0

    @property
    def path(self):
        return (
            f"{PACKAGEDIR}/data/sector{self.sector:03}/camera{self.camera:02}/ccd{self.ccd:02}/"
            f"tessbackdrop_sector{self.sector}_camera{self.camera}_ccd{self.ccd}.fits"
        )

    def _build_masks(self, frame):
        """Builds a set of pixel masks for the frame, which downweight saturated pixels or pixels with stars."""
        # if frame.shape != (2048, 2048):
        #     raise ValueError("Pass a frame that is (2048, 2048)")

        if not hasattr(self, "star_mask"):
            self.star_mask = get_star_mask(frame)

        if not hasattr(self, "sat_mask"):
            self.sat_mask = get_sat_mask(frame)
        if not hasattr(self, "jitter_mask"):
            if (~self.star_mask & self.sat_mask).sum() > self.njitter:
                s = np.random.choice(
                    (~self.star_mask & self.sat_mask).sum(),
                    size=self.njitter,
                    replace=False,
                )
                l = np.asarray(np.where(~self.star_mask & self.sat_mask))
                l = l[:, s]
                self.jitter_mask = np.zeros((self.cutout_size, self.cutout_size), bool)
                self.jitter_mask[l[0], l[1]] = True
            else:
                self.jitter_mask = (~self.star_mask & self.sat_mask).copy()

        if not hasattr(self, "bkg_mask"):
            if (self.star_mask & self.sat_mask).sum() > self.njitter:
                s = np.random.choice(
                    (self.star_mask & self.sat_mask).sum(),
                    size=self.njitter,
                    replace=False,
                )
                l = np.asarray(np.where(self.star_mask & self.sat_mask))
                l = l[:, s]
                self.bkg_mask = np.zeros((self.cutout_size, self.cutout_size), bool)
                self.bkg_mask[l[0], l[1]] = True
            else:
                self.bkg_mask = (self.star_mask & self.sat_mask).copy()

        sigma_f = xp.ones(frame.shape)
        sigma_f[~self.star_mask | ~self.sat_mask] = 1e6
        self.update_sigma_f(sigma_f)
        return

    @property
    def shape(self):
        if self.column is not None:
            return (self.row.shape[0], self.column.shape[0])
        else:
            return

    def _build_asteroid_mask(self, flux_cube, times):
        """Creates a mask for asteroids, and increases the flux errors in those
        pixels to reduce their contribution to the model

        Parameters
        ----------
        flux_cube: xp.ndarray or list of xp.ndarray
            3D flux cube of data. Be aware that this should be a short "batch"
            of flux data (e.g. 50 frames).
        """
        #        ast_mask = _asteroid_mask(flux_cube)
        ast_mask = get_asteroid_mask(
            sector=self.sector,
            camera=self.camera,
            ccd=self.ccd,
            cutout_size=flux_cube[0].shape[0],
            times=times,
        )
        sigma_f = xp.ones(flux_cube[0].shape)
        # More than the saturation limit
        sigma_f[~self.star_mask | ~self.sat_mask] = 1e6
        # Here we downweight asteroids and other variable pixels
        # The threshold is set very high, so we don't want these contributing at all
        sigma_f[ast_mask] = 1e6
        self.update_sigma_f(sigma_f)
        return ast_mask

    def _model_basic(self, tdx):
        """Model from the first design matrix"""
        return xp.power(10, self.A1.dot(self.weights_basic[tdx])).reshape(self.shape)

    def _model_full(self, tdx):
        """Model from the second design matrix"""
        return self.A2.dot(self.weights_full[tdx]).reshape(self.shape)

    def model(self, time_index=None):
        """Build a model for a frame with index `time_index`"""
        if time_index is None:
            time_index = np.arange(self.tstart.shape[0])
        single_frame = not hasattr(time_index, "__iter__")
        time_index = np.atleast_1d(time_index)
        if single_frame:
            return self._model_basic(time_index[0]) + self._model_full(time_index[0])
        return np.asarray(
            [self._model_basic(tdx) + self._model_full(tdx) for tdx in time_index]
        )

    def diagnose(self, fnames, ntimes=4):
        import matplotlib.pyplot as plt

        tdxs = np.linspace(0, self.weights_full.shape[0], ntimes + 2, dtype=int)[1:-1]
        fig, ax = plt.subplots(ntimes, 3, figsize=(4 * 3, 4 * ntimes))
        for jdx, tdx in enumerate(tdxs):
            while self.quality[tdx] != 0:
                tdx += 1
            dat = load_image(fnames[tdx])[self.row][:, self.column]
            v = np.nanpercentile(dat, (5, 95))
            ax[jdx, 0].imshow(dat, vmin=v[0], vmax=v[1])
            ax[jdx, 0].set_ylabel(f"{self.tstart[tdx]} BTJD")
            ax[jdx, 1].imshow(self.model(tdx), vmin=0, vmax=v[1] - v[0])
            ax[jdx, 2].imshow(dat - self.model(tdx), vmin=0, vmax=v[1] - v[0])
            for idx, title in enumerate(["Data", "Model", "Residuals"]):
                if jdx == 0:
                    ax[jdx, idx].set(title=title)
                ax[jdx, idx].set(xticks=[], yticks=[])
        plt.tight_layout()

    @property
    def average_frame(self):
        return (self._average_frame)[self.row, :][:, self.column]

    def _fit_basic_batch(self, flux):
        """Fit the first design matrix, in a batched mode"""
        # weights = list(self.A1.fit_batch(xp.log10(flux)))
        return self.A1.fit_batch(xp.log10(flux))

    def _fit_full_batch(self, flux):
        """Fit the second design matrix, in a batched mode"""
        #        weights = list(self.A2.fit_batch(flux))
        return self.A2.fit_batch(flux)

    def _fit_batch(self, flux_cube, times=None, mask_asteroids=True):
        """Fit the both design matrices, in a batched mode"""
        if mask_asteroids:
            if times is None:
                raise ValueError("Need times to mask asteroids")
            ast_mask = self._build_asteroid_mask(flux_cube, times=times)
        else:
            ast_mask = None
        weights_basic = self._fit_basic_batch(flux_cube)
        for tdx in range(len(weights_basic)):
            flux_cube[tdx] -= xp.power(10, self.A1.dot(weights_basic[tdx])).reshape(
                self.shape
            )
        weights_full = self._fit_full_batch(flux_cube)

        jitter, bkg = [], []
        for tdx in range(len(weights_full)):
            flux_cube[tdx] -= self.A2.dot(weights_full[tdx]).reshape(self.shape)
            jitter.append(flux_cube[tdx][self.jitter_mask])
            bkg.append(flux_cube[tdx][self.bkg_mask])
            # self._average_frame += flux_cube[tdx]
            # self._average_frame_count += 1
            flux_cube[tdx] += self.A2.dot(weights_full[tdx]).reshape(self.shape)

        for tdx in range(len(weights_basic)):
            flux_cube[tdx] += xp.power(10, self.A1.dot(weights_basic[tdx])).reshape(
                self.shape
            )
        return weights_basic, weights_full, jitter, bkg, ast_mask

    def _run_batches(self, input, tstart=None, batch_size=25, mask_asteroids=True):
        """
        Fit the model in batches, cutting the data into appropriate batch sizes.

        Parameters
        ----------
        input: list or np.ndarray
            list of filenames or 3D np.ndarray of flux values
        tstart: None or np.ndarray
            Array of tstart values, this is important for masking asteroids only
        batch_size: int
            Size to cut the data into in time
        mask_asteroids: bool
            Whether to mask out asteroids

        Returns
        -------
        weights_basic: np.ndarray
            The weights for the basic model. Has shape ntimes x nbasic weights
        weights_full: np.ndarray
            The weights for the basic model. Has shape ntimes x nfull weights
        jitter: np.ndarray
            The jitter components for the dataset, has size ntimes x njitter
        bkg: np.ndarray
            The background components for the dataset, has size ntimes x njitter
        asteroid_mask: np.ndarray
            A boolean mask showing where asteroids have been masked at some point
            in time. Has shape cutout_size x cutout_size.
        """
        if mask_asteroids and (tstart is None):
            raise ValueError("To mask asteroids, must pass `tstart` values.")

        nbatches = xp.ceil(len(input) / batch_size).astype(int)
        weights_basic, weights_full, jitter, bkg = [], [], [], []
        l = xp.arange(0, nbatches + 1, dtype=int) * batch_size
        if l[-1] > len(input):
            l[-1] = len(input)
        asteroid_mask = np.zeros((self.cutout_size, self.cutout_size), dtype=int)

        for l1, l2 in tqdm(
            zip(l[:-1], l[1:]),
            desc=f"{self.sector}, {self.camera}, {self.ccd} ",
            total=len(l) - 1,
            leave=True,
            position=0,
        ):
            if isinstance(input, (np.ndarray, list)) & isinstance(input[0], str):
                flux_cube = [
                    load_image(input[idx], cutout_size=self.cutout_size)
                    - self._average_frame
                    for idx in np.arange(l1, l2)
                ]
            elif isinstance(input, (np.ndarray, list)) & isinstance(
                input[0], np.ndarray
            ):
                flux_cube = [i - self._average_frame for i in input[l1:l2]]
            else:
                raise ValueError("Can not parse input.")

            if mask_asteroids:
                w1, w2, j, bk, ast_mask = self._fit_batch(
                    flux_cube,
                    times=tstart[l1:l2],
                    mask_asteroids=mask_asteroids,
                )
            else:
                w1, w2, j, bk, ast_mask = self._fit_batch(
                    flux_cube, mask_asteroids=mask_asteroids
                )
            weights_basic.append(w1)
            weights_full.append(w2)
            jitter.append(j)
            bkg.append(bk)
            if mask_asteroids:
                asteroid_mask += ast_mask

        return (
            xp.vstack(weights_basic),
            xp.vstack(weights_full),
            xp.vstack(jitter),
            xp.vstack(bkg),
            asteroid_mask,
        )

    def fit_model(
        self, flux_cube, tstart=None, batch_size=50, test_frame=0, mask_asteroids=False
    ):
        """Fit a model to a flux cube, fitting frames in batches of size `batch_size`.

        This will append to the list properties `self.weights_basic`, `self.weights_full`, `self.jitter`.

        Parameters
        ----------
        flux_cube : xp.ndarray
            3D array of frames.
        tstart: np.ndarray
            Times for each frame. Required if masking asteroids.
        batch_size : int
            Number of frames to fit at once.
        test_frame : int
            The index of the frame to use as the "reference frame".
            This reference frame will be used to build masks to set `sigma_f`.
            It should be the lowest background frame.
        """
        if isinstance(flux_cube, list):
            if not np.all(
                [f.shape == (self.cutout_size, self.cutout_size) for f in flux_cube]
            ):
                raise ValueError(
                    f"Frame is not ({self.cutout_size}, {self.cutout_size})"
                )
        elif isinstance(flux_cube, xp.ndarray):
            if flux_cube.ndim != 3:
                raise ValueError("`flux_cube` must be 3D")
            if not flux_cube.shape[1:] == (self.cutout_size, self.cutout_size):
                raise ValueError(
                    f"Frame is not ({self.cutout_size}, {self.cutout_size})"
                )
        else:
            raise ValueError("Pass an `xp.ndarray` or a `list`")
        if len(flux_cube) < 50:
            self._average_frame = np.zeros((self.cutout_size, self.cutout_size))
        else:
            self._average_frame = np.nanmin(flux_cube, axis=0)
        self._build_masks(flux_cube[test_frame] - self._average_frame)
        (
            self.weights_basic,
            self.weights_full,
            self.jitter,
            self.bkg,
            self.asteroid_mask,
        ) = self._run_batches(
            flux_cube,
            tstart=tstart,
            batch_size=batch_size,
            mask_asteroids=mask_asteroids,
        )
        # self._average_frame = flux_cube[test_frame] - self.model(test_frame)
        # # Mask out the asteroids from the average frame
        # if mask_asteroids:
        #     self._average_frame[
        #         get_asteroid_mask(
        #             self.sector,
        #             self.camera,
        #             self.ccd,
        #             cutout_size=self.cutout_size,
        #             times=[self.tstart[test_frame]],
        #         )
        #     ] = np.nanmedian(self._average_frame)
        # self._average_frame += min_frame

        _package_pca_comps(self)
        return

    def _package_weights_hdulist(self):
        hdu0 = self.hdu0
        s = np.argsort(self.tstart)
        cols = []
        for key in ["tstart", "tstop", "quality"]:
            if getattr(self, key) is not None:
                cols.append(
                    fits.Column(
                        name=key,
                        format="D",
                        unit="BJD - 2457000",
                        array=getattr(self, key)[s],
                    )
                )
        hdu1 = fits.BinTableHDU.from_columns(cols, name="TIME")
        hdu2 = fits.ImageHDU(np.asarray(self.weights_basic)[s], name="WEIGHTS1")
        hdu3 = fits.ImageHDU(np.asarray(self.weights_full)[s], name="WEIGHTS2")
        hdu4 = fits.ImageHDU(np.asarray(self.jitter_pack)[s], name="JITTER")
        hdu5 = fits.ImageHDU(np.asarray(self.bkg_pack)[s], name="BKG")
        hdu6 = fits.ImageHDU(np.asarray(self._average_frame), name="AVGFRAME")
        hdul = fits.HDUList([hdu0, hdu1, hdu2, hdu3, hdu4, hdu5, hdu6])
        return hdul

    def save(self, output_dir=None, overwrite=False):
        """Save a model fit to the scatterbrain data directory."""
        if not hasattr(self, "weights_basic"):
            raise ValueError("There is no model to save")
        self.hdu0 = fits.PrimaryHDU()
        self.hdu0.header["ORIGIN"] = "scatterbrain"
        self.hdu0.header["AUTHOR"] = "christina.l.hedges@nasa.gov"
        self.hdu0.header["VERSION"] = __version__
        for key in ["sector", "camera", "ccd"]:
            self.hdu0.header[key] = getattr(self, key)
        self.hdu0.header["CUTSIZE"] = getattr(self, "cutout_size")
        self.hdu0.header["STRPPOLY"] = getattr(self, "strap_npoly")

        if output_dir is None:
            output_dir = f"{PACKAGEDIR}/data/sector{self.sector:03}/camera{self.camera:02}/ccd{self.ccd:02}/"
            if output_dir != "":
                if not os.path.isdir(output_dir):
                    os.makedirs(output_dir)

        hdul = self._package_weights_hdulist()
        fname = (
            f"tessbackdrop_sector{self.sector}_camera{self.camera}_ccd{self.ccd}.fits"
        )
        hdul.writeto(output_dir + fname, overwrite=overwrite)

    def load(self, input, dir=None):
        """
        Load a model fit to the tess-backrop data directory.

        Parameters
        ----------
        input: tuple or string
            Either pass a tuple with `(sector, camera, ccd)` or pass
            a file name in `dir` to load
        dir : str
            Optional tring with the directory name

        Returns
        -------
        self: `scatterbrain.SimpleScatteredLightBackground` object
        """
        if isinstance(input, tuple):
            if len(input) == 3:
                sector, camera, ccd = input
                fname = f"tessbackdrop_sector{sector}_camera{camera}_ccd{ccd}.fits"
            else:
                raise ValueError("Please pass tuple as `(sector, camera, ccd)`")
        elif isinstance(input, str):
            fname = input
        else:
            raise ValueError("Can not parse input")
        if dir is None:
            dir = f"{PACKAGEDIR}/data/sector{sector:03}/camera{camera:02}/ccd{ccd:02}/"
        if dir != "":
            if not os.path.isdir(dir):
                raise ValueError("No solutions exist")
        if not os.path.isfile(dir + fname):
            if os.path.isfile(
                dir + f"sector{sector:03}/camera{camera:02}/ccd{ccd:02}/" + fname
            ):
                dir = dir + f"sector{sector:03}/camera{camera:02}/ccd{ccd:02}/"
            else:
                raise ValueError("No files exist")

        with fits.open(dir + fname, lazy_load_hdus=True) as hdu:
            for key in [
                "sector",
                "camera",
                "ccd",
            ]:
                setattr(self, key, hdu[0].header[key])
            setattr(self, "cutout_size", hdu[0].header["CUTSIZE"])

            if "tstart" in hdu[1].data.names:
                self.tstart = hdu[1].data["tstart"]
            if "tstop" in hdu[1].data.names:
                self.tstop = hdu[1].data["tstop"]
            if "quality" in hdu[1].data.names:
                self.quality = hdu[1].data["QUALITY"]
            self.weights_basic = list(hdu[2].data)
            weights_full = hdu[3].data
            self.strap_npoly = hdu[0].header["STRPPOLY"]
            # We have to do some work to get out just the column pixels
            # we are interested in
            strap_weights = np.hstack(
                (
                    weights_full[:, -2048 * self.strap_npoly :].reshape(
                        (weights_full.shape[0], self.strap_npoly, 2048)
                    )[:, :, self.column]
                ).transpose([1, 0, 2])
            )
            self.weights_full = np.hstack(
                [
                    weights_full[:, : -2048 * self.strap_npoly],
                    strap_weights,
                ]
            )
            self.jitter_pack = hdu[4].data
            self.bkg_pack = hdu[5].data
            # self._average_frame = hdu[6].data
            # self._average_frame_count = 1
            self._average_frame = hdu[6].data
        return self

    @staticmethod
    def from_disk(sector, camera, ccd, row=None, column=None, dir=None):
        return ScatteredLightBackground(
            sector=sector, camera=camera, ccd=ccd, row=row, column=column
        ).load((sector, camera, ccd), dir=dir)

    @staticmethod
    def from_tpf(tpf, dir=None):
        backdrop = ScatteredLightBackground(
            sector=tpf.sector,
            camera=tpf.camera,
            ccd=tpf.ccd,
            row=np.arange(tpf.shape[1]) + tpf.row - 1,
            column=np.arange(tpf.shape[2]) + tpf.column - 44 - 1,
        ).load((tpf.sector, tpf.camera, tpf.ccd), dir=dir)
        idx, jdx = _align_with_tpf(backdrop, tpf)
        return backdrop[idx]

    def get_tpf_mask(self, tpf):
        idx, jdx = _align_with_tpf(self, tpf)
        return jdx

    @staticmethod
    def from_tess_images(
        fnames,
        mask_asteroids=True,
        batch_size=50,
        test_frame=None,
        cutout_size=2048,
        sector=None,
        camera=None,
        ccd=None,
        verbose=False,
        quality_bitmask=175,
        njitter=5000,
    ):
        """Creates a ScatteredLightBackground model from filenames

        Parameters
        ----------
        fnames : list
            List of filenames to compute the background model for
        batch_size : int
            Number of files to process in a given batch
        test_frame : None or int
            The frame to use as a "test" frame for building masks.
            If None, a reasonable frame will be chosen.

        Returns
        -------
        b : `scatterbrain.ScatteredLightBackground.ScatteredLightBackground`
            ScatteredLightBackground object
        """
        fnames, sector, camera, ccd = _validate_inputs(fnames, sector, camera, ccd)

        blown_out_frames = identify_bad_frames(fnames)
        # make a ScatteredLightBackground object
        self = ScatteredLightBackground(
            sector=sector,
            camera=camera,
            ccd=ccd,
            cutout_size=cutout_size,
            njitter=njitter,
            verbose=verbose,
        )
        self.tstart = np.asarray(
            [fitsio.read_header(fname, ext=0)["TSTART"] for fname in fnames]
        )
        s = np.argsort(self.tstart)
        self.tstart, fnames = self.tstart[s], np.asarray(fnames)[s]
        self.tstop = np.asarray(
            [fitsio.read_header(fname, ext=0)["TSTOP"] for fname in fnames]
        )
        self.quality = np.asarray(
            [fitsio.read_header(fname, ext=1)["DQUALITY"] for fname in fnames]
        )
        self.quality[blown_out_frames] |= 8192
        # Step 1: find a good test frame
        if test_frame is None:
            a1, a2 = (
                np.min([self.A1.bore_pixel[0], 2047]),
                44 + np.min([self.A1.bore_pixel[1], 0]),
            )
            a1, a2 = a1.astype(int), a2.astype(int)

            # Test frame is a low flux frame, close to the middle of the dataset.
            f = np.asarray(
                [
                    fitsio.FITS(fname)[1][a1 : a1 + 1, a2 : a2 + 1][0][0]
                    for fname in fnames
                ]
            )
            l = np.where((f < np.nanpercentile(f, 5)) & (self.quality == 0))[0]
            if len(l) == 0:
                test_frame = len(fnames) // 2
            else:
                test_frame = l[np.argmin(np.abs(l - len(f) // 2))]
        ok = self.quality_mask(quality_bitmask)
        if ok.sum() < 50:
            self._average_frame = np.zeros((self.cutout_size, self.cutout_size))
        else:
            self._average_frame = (
                get_min_image_from_filenames(fnames[ok], cutout_size=self.cutout_size)
                - 1e-6
            )
        self._build_masks(
            load_image(fnames[test_frame], cutout_size=cutout_size)
            - self._average_frame
        )
        w1, w2, j, bk, self.asteroid_mask = self._run_batches(
            fnames[ok],
            self.tstart[ok],
            batch_size=batch_size,
            mask_asteroids=mask_asteroids,
        )

        self.weights_basic = np.zeros((len(fnames), w1.shape[1]))
        self.weights_basic[ok] = w1
        self.weights_basic[~ok] = np.nan
        self.weights_full = np.zeros((len(fnames), w2.shape[1]))
        self.weights_full[ok] = w2
        self.weights_full[~ok] = np.nan
        self.jitter = np.zeros((len(fnames), j.shape[1]))
        self.jitter[ok] = j
        self.jitter[~ok] = np.nan
        self.bkg = np.zeros((len(fnames), bk.shape[1]))
        self.bkg[ok] = bk
        self.bkg[~ok] = np.nan
        # self._average_frame = load_image(
        #     fnames[test_frame], cutout_size=cutout_size
        # ) - self.model(test_frame)
        # # Mask out the asteroids from the average frame
        # if mask_asteroids:
        #     self._average_frame[
        #         get_asteroid_mask(
        #             self.sector,
        #             self.camera,
        #             self.ccd,
        #             cutout_size=self.cutout_size,
        #             times=[self.tstart[test_frame]],
        #         )
        #     ] = np.nanmedian(self._average_frame)
        # self._average_frame += min_frame
        _package_pca_comps(self)
        return self



# flake8: noqa
import os

import fitsio
import numpy as np

try:
    if os.getenv("USE_CUPY") in ["True", "T", "true"]:
        import cupy as xp
        from cupy import sparse

        def load_image_cupy(fname, cutout_size=2048, loc=None):
            return xp.asarray(load_image_numpy(fname, cutout_size=cutout_size, loc=loc))

        def load_image(fname, cutout_size=2048, loc=None):
            return load_image_cupy(fname, cutout_size=cutout_size, loc=loc).astype(
                xp.float32
            )

    else:
        raise ImportError

except ImportError:
    import numpy as xp
    from scipy import sparse

    def load_image(fname, cutout_size=2048, loc=None):
        return load_image_numpy(fname, cutout_size=cutout_size, loc=loc).astype(
            xp.float32
        )


def load_image_numpy(fname, cutout_size=2048, loc=None):
    if loc is None:
        image = np.asarray(fitsio.FITS(fname)[1][:cutout_size, 44 : 44 + cutout_size])
    else:
        if not isinstance(loc, (np.ndarray, list)):
            raise ValueError("Pass a 2x2 location")
        if not np.asarray(loc).shape == (2, 2):
            raise ValueError("Pass a 2x2 location")
        if not (np.asarray(loc) <= 2048).all():
            raise ValueError("Pass a location on the detector")
        if not (np.asarray(loc).dtype == int):
            raise ValueError("Pass integers.")
        image = np.asarray(
            fitsio.FITS(fname)[1][
                loc[0][0] : loc[0][1], 44 + loc[1][0] : 44 + loc[1][1]
            ]
        )
    image[~np.isfinite(image)] = 1e-5
    image[image <= 0] = 1e-5
    return image



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['scatterbrain']

package_data = \
{'': ['*'],
 'scatterbrain': ['data/*',
                  'data/sector001/*',
                  'data/sector002/*',
                  'data/sector003/*',
                  'data/sector004/*',
                  'data/sector005/*',
                  'data/sector006/*',
                  'data/sector007/*',
                  'data/sector008/*',
                  'data/sector009/*',
                  'data/sector010/*',
                  'data/sector011/*',
                  'data/sector012/*',
                  'data/sector013/*',
                  'data/sector014/*',
                  'data/sector015/*',
                  'data/sector016/*',
                  'data/sector017/*',
                  'data/sector018/*',
                  'data/sector019/*',
                  'data/sector020/*',
                  'data/sector021/*',
                  'data/sector022/*',
                  'data/sector023/*',
                  'data/sector024/*',
                  'data/sector025/*',
                  'data/sector026/*']}

install_requires = \
['astropy>=4.3.1,<5.0.0',
 'fbpca>=1.0,<2.0',
 'fitsio>=1.1.5,<2.0.0',
 'matplotlib>=3.4.3,<4.0.0',
 'multiprocess>=0.70.12,<0.71.0',
 'numpy>=1.21.2,<2.0.0',
 'scipy>=1.7.1,<2.0.0',
 'tess-ephem>=0.3.0,<0.4.0',
 'tess-locator>=0.5.0,<0.6.0',
 'tqdm>=4.62.3,<5.0.0']

setup_kwargs = {
    'name': 'scatterbrain',
    'version': '0.1.3',
    'description': '',
    'long_description': None,
    'author': 'Christina Hedges',
    'author_email': 'christina.l.hedges@nasa.gov',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.10',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import

import os

PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

import pickle  # noqa

from .background import ScatteredLightBackground  # noqa
from .cupy_numpy_imports import load_image  # noqa
from .scene import StarScene  # noqa
from .tpf import correct_tpf  # noqa
from .version import __version__  # noqa

sector_times = pickle.load(open(f"{PACKAGEDIR}/data/tess_sector_times.pkl", "rb"))
